# -*- coding: utf-8 -*-

"""HDF5 input and output."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging

import numpy as np
import h5py
from six import string_types

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# HDF5 utility functions
#------------------------------------------------------------------------------

def _split_hdf5_path(path):
    """Return the group and dataset of the path."""
    # Make sure the path starts with a leading slash.
    if not path.startswith('/'):
        raise ValueError(("The HDF5 path '{0:s}' should start with a "
                          "leading slash '/'.").format(path))
    if '//' in path:
        raise ValueError(("There should be no double slash in the HDF5 path "
                          "'{0:s}'.").format(path))
    # Handle the special case '/'.
    if path == '/':
        return '/', ''
    # Temporarily remove the leading '/', we'll add it later (otherwise split
    # and join will mess it up).
    path = path[1:]
    # We split the path by slash and we get the head and tail.
    _split = path.split('/')
    group_path = '/'.join(_split[:-1])
    name = _split[-1]
    # Make some consistency checks.
    assert not group_path.endswith('/')
    assert '/' not in name
    # Finally, we add the leading slash at the beginning of the group path.
    return '/' + group_path, name


def _check_hdf5_path(h5_file, path):
    """Check that an HDF5 path exists in a file."""
    if path not in h5_file:
        raise ValueError("{path} doesn't exist.".format(path=path))


def _mmap_h5(path, h5path):
    with h5py.File(path, driver='sec2') as f:
        ds = f[h5path]
        offset = ds.id.get_offset()
        # Ensure we have a non-compressed contiguous array.
        assert ds.chunks is None
        assert ds.compression is None
        assert offset > 0
        dtype = ds.dtype
        shape = ds.shape
    arr = np.memmap(path, mode='r', shape=shape, offset=offset, dtype=dtype)
    return arr


#------------------------------------------------------------------------------
# File class
#------------------------------------------------------------------------------

class File(object):
    def __init__(self, filename, mode=None):
        if mode is None:
            mode = 'r'
        self.filename = filename
        self.mode = mode
        self._h5py_file = None

    # Open and close
    #--------------------------------------------------------------------------

    @property
    def h5py_file(self):
        """Native h5py file handle."""
        return self._h5py_file

    def is_open(self):
        return self._h5py_file is not None

    def open(self, mode=None):
        if mode is not None:
            self.mode = mode
        if not self.is_open():
            self._h5py_file = h5py.File(self.filename, self.mode)

    def close(self):
        if self.is_open():
            self._h5py_file.close()
            self._h5py_file = None

    # Datasets
    #--------------------------------------------------------------------------

    def read(self, path):
        """Read an HDF5 dataset, given its HDF5 path in the file."""
        _check_hdf5_path(self._h5py_file, path)
        return self._h5py_file[path]

    def write(self, path, array=None, dtype=None, shape=None, overwrite=False):
        """Write a NumPy array in the file.

        Parameters
        ----------
        path : str
            Full HDF5 path to the dataset to create.
        array : ndarray
            Array to write in the file.
        dtype : dtype
            If `array` is None, the dtype of the array.
        shape : tuple
            If `array` is None, the shape of the array.
        overwrite : bool
            If False, raise an error if the dataset already exists. Defaults
            to False.

        """
        # Get the group path and the dataset name.
        group_path, dset_name = _split_hdf5_path(path)

        # If the parent group doesn't already exist, create it.
        if group_path not in self._h5py_file:
            self._h5py_file.create_group(group_path)

        group = self._h5py_file[group_path]

        # Check that the dataset does not already exist.
        if path in self._h5py_file:
            if overwrite:
                # Force rewriting the dataset if 'overwrite' is True.
                del self._h5py_file[path]
            else:
                # Otherwise, raise an error.
                raise ValueError(("The dataset '{0:s}' already exists."
                                  ).format(path))

        if array is not None:
            return group.create_dataset(dset_name, data=array)
        else:
            assert dtype
            assert shape
            return group.create_dataset(dset_name, dtype=dtype, shape=shape)

    # Copy and rename
    #--------------------------------------------------------------------------

    def _check_move_copy(self, path, new_path):
        if not self.exists(path):
            raise ValueError("'{0}' doesn't exist.".format(path))
        if self.exists(new_path):
            raise ValueError("'{0}' already exist.".format(new_path))

    def move(self, path, new_path):
        """Move a group or dataset to another location."""
        self._check_move_copy(path, new_path)
        self._h5py_file.move(path, new_path)

    def copy(self, path, new_path):
        """Copy a group or dataset to another location."""
        self._check_move_copy(path, new_path)
        self._h5py_file.copy(path, new_path)

    def delete(self, path):
        """Delete a group or dataset."""
        if not path.startswith('/'):
            path = '/' + path
        if not self.exists(path):
            raise ValueError("'{0}' doesn't exist.".format(path))

        path, name = _split_hdf5_path(path)
        parent = self.read(path)
        del parent[name]

    # Attributes
    #--------------------------------------------------------------------------

    def read_attr(self, path, attr_name, default=False):
        """Read an attribute of an HDF5 group."""
        _check_hdf5_path(self._h5py_file, path)
        attrs = self._h5py_file[path].attrs
        if attr_name in attrs:
            try:
                out = attrs[attr_name]
                if (isinstance(out, (np.ndarray, np.generic)) and
                        out.dtype.kind == 'S'):
                    out = out.tostring().decode('UTF-8')
                    out = out.replace('\x00', '')
                return out
            except (TypeError, IOError):
                logger.debug("Unable to read attribute `%s` at `%s`.",
                             attr_name, path)
                return
        elif default is False:
            raise KeyError("The attribute '{0:s}' ".format(attr_name) +
                           "at `{}` doesn't exist.".format(path))
        return default

    def write_attr(self, path, attr_name, value):
        """Write an attribute of an HDF5 group."""
        assert isinstance(path, string_types)
        assert isinstance(attr_name, string_types)
        if value is None:
            value = ''
        # Ensure lists of strings are converted to ASCII arrays.
        if isinstance(value, list):
            if not value:
                value = None
            if value and isinstance(value[0], string_types):
                value = np.array(value, dtype='S')
        # Use string arrays instead of vlen arrays (crash in h5py 2.5.0 win64).
        if isinstance(value, string_types):
            value = np.array([value], dtype='S')
        # Idem: fix crash with boolean attributes on win64.
        if isinstance(value, bool):
            value = int(value)
        # If the parent group doesn't already exist, create it.
        if path not in self._h5py_file:
            self._h5py_file.create_group(path)
        try:
            self._h5py_file[path].attrs[attr_name] = value
            # debug("Write `{}={}` at `{}`.".format(attr_name,
            #                                       str(value), path))
        except TypeError:
            logger.warn("Unable to write attribute `%s=%s` at `%s`.",
                        attr_name, value, path)

    def attrs(self, path='/'):
        """Return the list of attributes at the given path."""
        if path in self._h5py_file:
            return sorted(self._h5py_file[path].attrs)
        else:
            return []

    def has_attr(self, path, attr_name):
        """Return whether an attribute exists at a given path."""
        if path not in self._h5py_file:
            return False
        else:
            return attr_name in self._h5py_file[path].attrs

    # Children
    #--------------------------------------------------------------------------

    def children(self, path='/'):
        """Return the list of children of a given node."""
        return sorted(self._h5py_file[path].keys())

    def groups(self, path='/'):
        """Return the list of groups under a given node."""
        return [key for key in self.children(path)
                if isinstance(self._h5py_file[path + '/' + key],
                              h5py.Group)]

    def datasets(self, path='/'):
        """Return the list of datasets under a given node."""
        return [key for key in self.children(path)
                if isinstance(self._h5py_file[path + '/' + key],
                              h5py.Dataset)]

    # Miscellaneous properties
    #--------------------------------------------------------------------------

    def exists(self, path):
        return path in self._h5py_file

    def _print_node_info(self, name, node):
        """Print node information."""
        info = ('/' + name).ljust(50)
        if isinstance(node, h5py.Group):
            pass
        elif isinstance(node, h5py.Dataset):
            info += str(node.shape).ljust(20)
            info += str(node.dtype).ljust(8)
        print(info)

    def describe(self):
        """Display the list of all groups and datasets in the file."""
        if not self.is_open():
            raise IOError("Cannot display file information because the file"
                          " '{0:s}' is not open.".format(self.filename))
        self._h5py_file['/'].visititems(self._print_node_info)

    # Context manager
    #--------------------------------------------------------------------------

    def __contains__(self, path):
        return path in self._h5py_file

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, tb):
        self.close()


def open_h5(filename, mode=None):
    """Open an HDF5 file and return a File instance."""
    file = File(filename, mode=mode)
    file.open()
    return file



# -*- coding: utf-8 -*-

"""PCA for features."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np


#------------------------------------------------------------------------------
# PCA
#------------------------------------------------------------------------------

def _compute_pcs(x, n_pcs=None, masks=None):
    """Compute the PCs of waveforms."""

    assert x.ndim == 3
    x = np.asarray(x, np.float64)

    n_spikes, n_samples, n_channels = x.shape

    if masks is not None:
        assert isinstance(masks, np.ndarray)
        assert masks.shape == (n_spikes, n_channels)

    # Compute regularization cov matrix.
    cov_reg = np.eye(n_samples)
    if masks is not None:
        unmasked = masks > 0
        # The last dimension is now time. The second dimension is channel.
        x_swapped = np.swapaxes(x, 1, 2)
        # This is the list of all unmasked spikes on all channels.
        # shape: (n_unmasked_spikes, n_samples)
        unmasked_all = x_swapped[unmasked, :]
        # Let's compute the regularization cov matrix of this beast.
        # shape: (n_samples, n_samples)
        cov_reg_ = np.cov(unmasked_all, rowvar=0)
        # Make sure the covariance matrix is valid.
        if cov_reg_.ndim == 2:
            cov_reg = cov_reg
    assert cov_reg.shape == (n_samples, n_samples)

    pcs_list = []
    # Loop over channels
    for channel in range(n_channels):
        x_channel = x[:, :, channel]
        # Compute cov matrix for the channel
        if masks is not None:
            # Unmasked waveforms on that channel
            # shape: (n_unmasked, n_samples)
            x_channel = np.compress(masks[:, channel] > 0,
                                    x_channel, axis=0)
        assert x_channel.ndim == 2
        # Don't compute the cov matrix if there are no unmasked spikes
        # on that channel.
        alpha = 1. / n_spikes
        if x_channel.shape[0] <= 1:
            cov = alpha * cov_reg
        else:
            cov_channel = np.cov(x_channel, rowvar=0)
            assert cov_channel.shape == (n_samples, n_samples)
            cov = alpha * cov_reg + cov_channel
        # Compute the eigenelements
        vals, vecs = np.linalg.eigh(cov)
        pcs = vecs.T.astype(np.float32)[np.argsort(vals)[::-1]]
        # Take the first n_pcs components.
        if n_pcs is not None:
            pcs = pcs[:n_pcs, ...]
        pcs_list.append(pcs[:n_pcs, ...])

    pcs = np.dstack(pcs_list)
    return pcs


def _project_pcs(x, pcs):
    """Project data points onto principal components.

    Parameters
    ----------

    x : array
        The waveforms
    pcs : array
        The PCs returned by `_compute_pcs()`.
    """
    # pcs: (nf, ns, nc)
    # x: (n, ns, nc)
    # out: (n, nc, nf)
    assert pcs.ndim == 3
    assert x.ndim == 3
    n, ns, nc = x.shape
    nf, ns_, nc_ = pcs.shape
    assert ns == ns_
    assert nc == nc_

    x_proj = np.einsum('...ijk,...jk->...ki', pcs, x)
    assert x_proj.shape == (n, nc, nf)
    return x_proj


class PCA(object):
    """Apply PCA to waveforms."""
    def __init__(self, n_pcs=None):
        self._n_pcs = n_pcs
        self._pcs = None

    def fit(self, waveforms, masks=None):
        """Compute the PCs of waveforms.

        Parameters
        ----------

        waveforms : ndarray
            Shape: `(n_spikes, n_samples, n_channels)`
        masks : ndarray
            Shape: `(n_spikes, n_channels)`

        """
        self._pcs = _compute_pcs(waveforms, n_pcs=self._n_pcs, masks=masks)
        return self._pcs

    def transform(self, waveforms, pcs=None):
        """Project waveforms on the PCs.

        Parameters
        ----------

        waveforms : ndarray
            Shape: `(n_spikes, n_samples, n_channels)`

        """
        if pcs is None:
            pcs = self._pcs
        # Need to call fit() if the pcs are None here.
        if pcs is not None:
            return _project_pcs(waveforms, pcs)



# -*- coding: utf-8 -*-

"""Multi-electrode arrays."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os
import os.path as op
import itertools

import numpy as np

from ..utils import _read_python


#------------------------------------------------------------------------------
# PRB file utilities
#------------------------------------------------------------------------------

def _edges_to_adjacency_list(edges):
    """Convert a list of edges into an adjacency list."""
    adj = {}
    for i, j in edges:
        if i in adj:
            ni = adj[i]
        else:
            ni = adj[i] = set()
        if j in adj:
            nj = adj[j]
        else:
            nj = adj[j] = set()
        ni.add(j)
        nj.add(i)
    return adj


def _probe_positions(probe, group):
    """Return the positions of a probe channel group."""
    positions = probe['channel_groups'][group]['geometry']
    channels = _probe_channels(probe, group)
    return np.array([positions[channel] for channel in channels])


def _probe_channels(probe, group):
    """Return the list of channels in a channel group.

    The order is kept.

    """
    return probe['channel_groups'][group]['channels']


def _probe_all_channels(probe):
    """Return the list of channels in the probe."""
    cgs = probe['channel_groups'].values()
    cg_channels = [cg['channels'] for cg in cgs]
    return sorted(set(itertools.chain(*cg_channels)))


def _probe_adjacency_list(probe):
    """Return an adjacency list of a whole probe."""
    cgs = probe['channel_groups'].values()
    graphs = [cg['graph'] for cg in cgs]
    edges = list(itertools.chain(*graphs))
    adjacency_list = _edges_to_adjacency_list(edges)
    return adjacency_list


def _channels_per_group(probe):
    groups = probe['channel_groups'].keys()
    return {group: probe['channel_groups'][group]['channels']
            for group in groups}


def load_probe(name):
    """Load one of the built-in probes."""
    if op.exists(name):
        # The argument can be either a path to a PRB file.
        path = name
    else:
        # Or the name of a built-in probe.
        curdir = op.realpath(op.dirname(__file__))
        path = op.join(curdir, '../probes/{}.prb'.format(name))
    if not op.exists(path):
        raise IOError("The probe `{}` cannot be found.".format(name))
    return _read_python(path)


def list_probes():
    """Return the list of built-in probes."""
    curdir = op.realpath(op.dirname(__file__))
    return [op.splitext(fn)[0]
            for fn in os.listdir(op.join(curdir, '../probes'))
            if fn.endswith('.prb')]


#------------------------------------------------------------------------------
# MEA class
#------------------------------------------------------------------------------

class MEA(object):
    """A Multi-Electrode Array.

    There are two modes:

    * No probe specified: one single channel group, positions and adjacency
      list specified directly.
    * Probe specified: one can change the current channel_group.

    """

    def __init__(self,
                 channels=None,
                 positions=None,
                 adjacency=None,
                 probe=None,
                 ):
        self._probe = probe
        self._channels = channels
        if positions is not None:
            assert self.n_channels == positions.shape[0]
        self._positions = positions
        # This is a mapping {channel: list of neighbors}.
        if adjacency is None and probe is not None:
            adjacency = _probe_adjacency_list(probe)
            self.channels_per_group = _channels_per_group(probe)
        self._adjacency = adjacency

    def _check_positions(self, positions):
        if positions is None:
            return
        positions = np.asarray(positions)
        if self.n_channels is None:
            self.n_channels = positions.shape[0]
        if positions.shape[0] != self.n_channels:
            raise ValueError("'positions' "
                             "(shape {0:s})".format(str(positions.shape)) +
                             " and 'n_channels' "
                             "({0:d})".format(self.n_channels) +
                             " do not match.")

    @property
    def positions(self):
        """Channel positions in the current channel group."""
        return self._positions

    @positions.setter
    def positions(self, value):
        self._check_positions(value)
        self._positions = value

    @property
    def channels(self):
        """Channel ids in the current channel group."""
        return self._channels

    @property
    def n_channels(self):
        """Number of channels in the current channel group."""
        return len(self._channels) if self._channels is not None else 0

    @property
    def adjacency(self):
        """Adjacency graph in the current channel group."""
        return self._adjacency

    @adjacency.setter
    def adjacency(self, value):
        self._adjacency = value

    def change_channel_group(self, group):
        """Change the current channel group."""
        assert self._probe is not None
        self._channels = _probe_channels(self._probe, group)
        self._positions = _probe_positions(self._probe, group)


#------------------------------------------------------------------------------
# Common probes
#------------------------------------------------------------------------------

def linear_positions(n_channels):
    """Linear channel positions along the vertical axis."""
    return np.c_[np.zeros(n_channels),
                 np.linspace(0., 1., n_channels)]


def staggered_positions(n_channels):
    """Generate channel positions for a staggered probe."""
    i = np.arange(n_channels - 1)
    x, y = (-1) ** i * (5 + i), 10 * (i + 1)
    pos = np.flipud(np.r_[np.zeros((1, 2)), np.c_[x, y]])
    return pos



# -*- coding: utf-8 -*-

"""Tests of misc type utility functions."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op

import numpy as np
from numpy.testing import assert_array_equal as ae
from pytest import raises, mark

from ..kwik.mock import artificial_spike_clusters
from ..utils import (Bunch, _is_integer, captured_output,
                     _load_arrays, _save_arrays,
                     _pad,
                     _spikes_per_cluster,
                     get_excerpts,
                     chunk_bounds,
                     data_chunk,
                     excerpts,
                     )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_bunch():
    obj = Bunch()
    obj['a'] = 1
    assert obj.a == 1
    obj.b = 2
    assert obj['b'] == 2


def test_integer():
    assert _is_integer(3)
    assert _is_integer(np.arange(1)[0])
    assert not _is_integer(3.)


def test_captured_output():
    with captured_output() as (out, err):
        print('Hello world!')
    assert out.getvalue().strip() == 'Hello world!'


@mark.parametrize('n', [20, 0])
def test_load_save_arrays(tempdir, n):
    path = op.join(tempdir, 'test.npy')
    # Random arrays.
    arrays = []
    for i in range(n):
        size = np.random.randint(low=3, high=50)
        assert size > 0
        arr = np.random.rand(size, 3).astype(np.float32)
        arrays.append(arr)
    _save_arrays(path, arrays)

    arrays_loaded = _load_arrays(path)

    assert len(arrays) == len(arrays_loaded)
    for arr, arr_loaded in zip(arrays, arrays_loaded):
        assert arr.shape == arr_loaded.shape
        assert arr.dtype == arr_loaded.dtype
        ae(arr, arr_loaded)


def test_pad():
    arr = np.random.rand(10, 3)

    ae(_pad(arr, 0, 'right'), arr[:0, :])
    ae(_pad(arr, 3, 'right'), arr[:3, :])
    ae(_pad(arr, 9), arr[:9, :])
    ae(_pad(arr, 10), arr)

    ae(_pad(arr, 12, 'right')[:10, :], arr)
    ae(_pad(arr, 12)[10:, :], np.zeros((2, 3)))

    ae(_pad(arr, 0, 'left'), arr[:0, :])
    ae(_pad(arr, 3, 'left'), arr[7:, :])
    ae(_pad(arr, 9, 'left'), arr[1:, :])
    ae(_pad(arr, 10, 'left'), arr)

    ae(_pad(arr, 12, 'left')[2:, :], arr)
    ae(_pad(arr, 12, 'left')[:2, :], np.zeros((2, 3)))

    with raises(ValueError):
        _pad(arr, -1)


def test_spikes_per_cluster():
    """Test _spikes_per_cluster()."""

    n_spikes = 1000
    spike_ids = np.arange(n_spikes).astype(np.int64)
    n_clusters = 10
    spike_clusters = artificial_spike_clusters(n_spikes, n_clusters)

    spikes_per_cluster = _spikes_per_cluster(spike_ids, spike_clusters)
    assert list(spikes_per_cluster.keys()) == list(range(n_clusters))

    for i in range(10):
        ae(spikes_per_cluster[i], np.sort(spikes_per_cluster[i]))
        assert np.all(spike_clusters[spikes_per_cluster[i]] == i)


#------------------------------------------------------------------------------
# Test chunking
#------------------------------------------------------------------------------

def test_chunk_bounds():
    chunks = chunk_bounds(200, 100, overlap=20)

    assert next(chunks) == (0, 100, 0, 90)
    assert next(chunks) == (80, 180, 90, 170)
    assert next(chunks) == (160, 200, 170, 200)


def test_chunk():
    data = np.random.randn(200, 4)
    chunks = chunk_bounds(data.shape[0], 100, overlap=20)

    with raises(ValueError):
        data_chunk(data, (0, 0, 0))

    assert data_chunk(data, (0, 0)).shape == (0, 4)

    # Chunk 1.
    ch = next(chunks)
    d = data_chunk(data, ch)
    d_o = data_chunk(data, ch, with_overlap=True)

    ae(d_o, data[0:100])
    ae(d, data[0:90])

    # Chunk 2.
    ch = next(chunks)
    d = data_chunk(data, ch)
    d_o = data_chunk(data, ch, with_overlap=True)

    ae(d_o, data[80:180])
    ae(d, data[90:170])


def test_excerpts_1():
    bounds = [(start, end) for (start, end) in excerpts(100,
                                                        n_excerpts=3,
                                                        excerpt_size=10)]
    assert bounds == [(0, 10), (45, 55), (90, 100)]


def test_excerpts_2():
    bounds = [(start, end) for (start, end) in excerpts(10,
                                                        n_excerpts=3,
                                                        excerpt_size=10)]
    assert bounds == [(0, 10)]


def test_get_excerpts():
    data = np.random.rand(100, 2)
    subdata = get_excerpts(data, n_excerpts=10, excerpt_size=5)
    assert subdata.shape == (50, 2)
    ae(subdata[:5, :], data[:5, :])
    ae(subdata[-5:, :], data[-10:-5, :])

    data = np.random.rand(10, 2)
    subdata = get_excerpts(data, n_excerpts=10, excerpt_size=5)
    ae(subdata, data)



# -*- coding: utf-8 -*-
# flake8: noqa

"""Spike detection, waveform extraction."""

from .detect import Thresholder, FloodFillDetector, compute_threshold
from .filter import Filter, Whitening
from .pca import PCA
from .spikedetekt import SpikeDetekt
from .waveform import WaveformLoader, WaveformExtractor, SpikeLoader



# -*- coding: utf-8 -*-

"""Mock Kwik files."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op

import numpy as np
import numpy.random as nr

from .mea import staggered_positions
from .h5 import open_h5
from .model import _create_clustering


#------------------------------------------------------------------------------
# Mock functions
#------------------------------------------------------------------------------

def artificial_waveforms(n_spikes=None, n_samples=None, n_channels=None):
    # TODO: more realistic waveforms.
    return .25 * nr.normal(size=(n_spikes, n_samples, n_channels))


def artificial_features(*args):
    return .25 * nr.normal(size=args)


def artificial_masks(n_spikes=None, n_channels=None):
    masks = nr.uniform(size=(n_spikes, n_channels))
    masks[masks < .25] = 0
    return masks


def artificial_traces(n_samples, n_channels):
    # TODO: more realistic traces.
    return .25 * nr.normal(size=(n_samples, n_channels))


def artificial_spike_clusters(n_spikes, n_clusters, low=0):
    return nr.randint(size=n_spikes, low=low, high=max(1, n_clusters))


def artificial_spike_samples(n_spikes, max_isi=50):
    return np.cumsum(nr.randint(low=0, high=max_isi, size=n_spikes))


def artificial_correlograms(n_clusters, n_samples):
    return nr.uniform(size=(n_clusters, n_clusters, n_samples))


def mock_prm(dat_path):
    return dict(
        prb_file='1x32_buzsaki',
        traces=dict(
            raw_data_files=[dat_path, dat_path],
            voltage_gain=10.,
            sample_rate=20000,
            n_channels=32,
            dtype='int16',
        ),
        spikedetekt={
            'n_features_per_channel': 4,
        },
        klustakwik2={},
    )


#------------------------------------------------------------------------------
# Mock Kwik file
#------------------------------------------------------------------------------

def create_mock_kwik(dir_path, n_clusters=None, n_spikes=None,
                     n_channels=None, n_features_per_channel=None,
                     n_samples_traces=None,
                     with_kwx=True,
                     with_kwd=True,
                     add_original=True,
                     ):
    """Create a test kwik file."""
    filename = op.join(dir_path, '_test.kwik')
    kwx_filename = op.join(dir_path, '_test.kwx')
    kwd_filename = op.join(dir_path, '_test.raw.kwd')

    # Create the kwik file.
    with open_h5(filename, 'w') as f:
        f.write_attr('/', 'kwik_version', 2)

        def _write_metadata(key, value):
            f.write_attr('/application_data/spikedetekt', key, value)

        _write_metadata('sample_rate', 20000.)

        # Filter parameters.
        _write_metadata('filter_low', 500.)
        _write_metadata('filter_high_factor', 0.95 * .5)
        _write_metadata('filter_butter_order', 3)

        _write_metadata('extract_s_before', 15)
        _write_metadata('extract_s_after', 25)

        _write_metadata('n_features_per_channel', n_features_per_channel)

        # Create spike times.
        spike_samples = artificial_spike_samples(n_spikes).astype(np.int64)
        spike_recordings = np.zeros(n_spikes, dtype=np.uint16)
        # Size of the first recording.
        recording_size = 2 * n_spikes // 3
        if recording_size > 0:
            # Find the recording offset.
            recording_offset = spike_samples[recording_size]
            recording_offset += spike_samples[recording_size + 1]
            recording_offset //= 2
            spike_recordings[recording_size:] = 1
            # Make sure the spike samples of the second recording start over.
            spike_samples[recording_size:] -= spike_samples[recording_size]
            spike_samples[recording_size:] += 10
        else:
            recording_offset = 1

        if spike_samples.max() >= n_samples_traces:
            raise ValueError("There are too many spikes: decrease 'n_spikes'.")

        f.write('/channel_groups/1/spikes/time_samples', spike_samples)
        f.write('/channel_groups/1/spikes/recording', spike_recordings)
        f.write_attr('/channel_groups/1', 'channel_order',
                     np.arange(1, n_channels - 1)[::-1])
        graph = np.array([[1, 2], [2, 3]])
        f.write_attr('/channel_groups/1', 'adjacency_graph', graph)

        # Create channels.
        positions = staggered_positions(n_channels)
        for channel in range(n_channels):
            group = '/channel_groups/1/channels/{0:d}'.format(channel)
            f.write_attr(group, 'name', str(channel))
            f.write_attr(group, 'position', positions[channel])

        # Create spike clusters.
        clusterings = [('main', n_clusters)]
        if add_original:
            clusterings += [('original', n_clusters * 2)]
        for clustering, n_clusters_rec in clusterings:
            spike_clusters = artificial_spike_clusters(n_spikes,
                                                       n_clusters_rec)
            groups = {0: 0, 1: 1, 2: 2}
            _create_clustering(f, clustering, 1, spike_clusters, groups)

        # Create recordings.
        f.write_attr('/recordings/0', 'name', 'recording_0')
        f.write_attr('/recordings/1', 'name', 'recording_1')

        f.write_attr('/recordings/0/raw', 'hdf5_path', kwd_filename)
        f.write_attr('/recordings/1/raw', 'hdf5_path', kwd_filename)

    # Create the kwx file.
    if with_kwx:
        with open_h5(kwx_filename, 'w') as f:
            f.write_attr('/', 'kwik_version', 2)
            features = artificial_features(n_spikes,
                                           (n_channels - 2) *
                                           n_features_per_channel)
            masks = artificial_masks(n_spikes,
                                     (n_channels - 2) *
                                     n_features_per_channel)
            fm = np.dstack((features, masks)).astype(np.float32)
            f.write('/channel_groups/1/features_masks', fm)

    # Create the raw kwd file.
    if with_kwd:
        with open_h5(kwd_filename, 'w') as f:
            f.write_attr('/', 'kwik_version', 2)
            traces = artificial_traces(n_samples_traces, n_channels)
            # TODO: int16 traces
            f.write('/recordings/0/data',
                    traces[:recording_offset, ...].astype(np.float32))
            f.write('/recordings/1/data',
                    traces[recording_offset:, ...].astype(np.float32))

    return filename



# -*- coding: utf-8 -*-

"""Wrapper to KlustaKwik2 implementation."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
import six

from .utils import chunk_bounds
from .kwik.model import PartialArray


#------------------------------------------------------------------------------
# Clustering class
#------------------------------------------------------------------------------

def sparsify_features_masks(features, masks, chunk_size=10000):
    from klustakwik2 import RawSparseData

    assert features.ndim == 2
    assert masks.ndim == 2
    assert features.shape == masks.shape

    n_spikes, num_features = features.shape

    # Stage 1: read min/max of fet values for normalization
    # and count total number of unmasked features.
    vmin = np.ones(num_features) * np.inf
    vmax = np.ones(num_features) * (-np.inf)
    total_unmasked_features = 0
    for _, _, i, j in chunk_bounds(n_spikes, chunk_size):
        f, m = features[i:j], masks[i:j]
        inds = m > 0
        # Replace the masked values by NaN.
        vmin = np.minimum(np.min(f, axis=0), vmin)
        vmax = np.maximum(np.max(f, axis=0), vmax)
        total_unmasked_features += inds.sum()
    # Stage 2: read data line by line, normalising
    vdiff = vmax - vmin
    vdiff[vdiff == 0] = 1
    fetsum = np.zeros(num_features)
    fet2sum = np.zeros(num_features)
    nsum = np.zeros(num_features)
    all_features = np.zeros(total_unmasked_features)
    all_fmasks = np.zeros(total_unmasked_features)
    all_unmasked = np.zeros(total_unmasked_features, dtype=int)
    offsets = np.zeros(n_spikes + 1, dtype=int)
    curoff = 0
    for i in six.moves.range(n_spikes):
        fetvals, fmaskvals = (features[i] - vmin) / vdiff, masks[i]
        inds = (fmaskvals > 0).nonzero()[0]
        masked_inds = (fmaskvals == 0).nonzero()[0]
        all_features[curoff:curoff + len(inds)] = fetvals[inds]
        all_fmasks[curoff:curoff + len(inds)] = fmaskvals[inds]
        all_unmasked[curoff:curoff + len(inds)] = inds
        offsets[i] = curoff
        curoff += len(inds)
        fetsum[masked_inds] += fetvals[masked_inds]
        fet2sum[masked_inds] += fetvals[masked_inds] ** 2
        nsum[masked_inds] += 1
    offsets[-1] = curoff

    nsum[nsum == 0] = 1
    noise_mean = fetsum / nsum
    noise_variance = fet2sum / nsum - noise_mean ** 2

    return RawSparseData(noise_mean,
                         noise_variance,
                         all_features,
                         all_fmasks,
                         all_unmasked,
                         offsets,
                         )


def klustakwik(model=None,
               spike_ids=None,
               spike_clusters=None,  # Initial spike clusters.
               features=None,
               masks=None,
               iter_callback=None,
               **kwargs
               ):
    """Run the clustering algorithm on the model, or on any features
    and masks.

    Return the `spike_clusters` assignements.

    """
    # Get the features and masks.
    if model is not None:
        if features is None:
            features = PartialArray(model.all_features_masks, 0)
        if masks is None:
            masks = PartialArray(model.all_features_masks, 1)
    # Select some spikes if needed.
    if spike_ids is not None:
        features = features[spike_ids]
        masks = masks[spike_ids]
    # Convert the features and masks to the sparse structure used
    # by KK.
    data = sparsify_features_masks(features, masks)
    data = data.to_sparse_data()

    # Instantiate KK.
    from klustakwik2 import KK, __version__
    kk = KK(data, **kwargs)

    @kk.register_callback
    def f(_):
        # Skip split iterations.
        if _.name != '':
            return
        if iter_callback:
            iter_callback(kk.clusters)

    # Launch KK.
    if spike_clusters is None:
        kk.cluster_mask_starts()
    else:
        kk.cluster_from(spike_clusters)
    spike_clusters = kk.clusters
    params = kk.params
    params['version'] = __version__
    return spike_clusters, params



# -*- coding: utf-8 -*-

"""Tests of dataset utility functions."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op
from itertools import product

import numpy as np
from numpy.testing import assert_array_equal as ae
import responses
from pytest import raises, yield_fixture

from ..datasets import (download_file,
                        download_test_data,
                        download_sample_data,
                        _check_md5_of_url,
                        _BASE_URL,
                        )


#------------------------------------------------------------------------------
# Fixtures
#------------------------------------------------------------------------------

# Test URL and data
_URL = 'http://test/data'
_DATA = np.linspace(0., 1., 100000).astype(np.float32)
_CHECKSUM = '7d257d0ae7e3af8ca3574ccc3a4bf072'


def _add_mock_response(url, body, file_type='binary'):
    content_type = ('application/octet-stream'
                    if file_type == 'binary' else 'text/plain')
    responses.add(responses.GET, url,
                  body=body,
                  status=200,
                  content_type=content_type,
                  )


@yield_fixture
def mock_url():
    _add_mock_response(_URL, _DATA.tostring())
    _add_mock_response(_URL + '.md5', _CHECKSUM + '  ' + op.basename(_URL))
    yield
    responses.reset()


@yield_fixture(params=product((True, False), repeat=4))
def mock_urls(request):
    data = _DATA.tostring()
    checksum = _CHECKSUM
    url_data = _URL
    url_checksum = _URL + '.md5'

    if not request.param[0]:
        # Data URL is corrupted.
        url_data = url_data[:-1]
    if not request.param[1]:
        # Data is corrupted.
        data = data[:-1]
    if not request.param[2]:
        # Checksum URL is corrupted.
        url_checksum = url_checksum[:-1]
    if not request.param[3]:
        # Checksum is corrupted.
        checksum = checksum[:-1]

    _add_mock_response(url_data, data)
    _add_mock_response(url_checksum, checksum)
    yield request.param, url_data, url_checksum
    responses.reset()


def _dl(path):
    download_file(_URL, path)
    with open(path, 'rb') as f:
        data = f.read()
    return data


def _check(data):
    ae(np.fromstring(data, np.float32), _DATA)


#------------------------------------------------------------------------------
# Test utility functions
#------------------------------------------------------------------------------

@responses.activate
def test_check_md5_of_url(tempdir, mock_url):
    output_path = op.join(tempdir, 'data')
    download_file(_URL, output_path)
    assert _check_md5_of_url(output_path, _URL)


#------------------------------------------------------------------------------
# Test download functions
#------------------------------------------------------------------------------

@responses.activate
def test_download_not_found(tempdir):
    path = op.join(tempdir, 'test')
    with raises(Exception):
        download_file(_URL + '_notfound', path)


@responses.activate
def test_download_already_exists_invalid(tempdir, mock_url):
    path = op.join(tempdir, 'test')
    # Create empty file.
    open(path, 'a').close()
    _check(_dl(path))


@responses.activate
def test_download_already_exists_valid(tempdir, mock_url):
    path = op.join(tempdir, 'test')
    # Create valid file.
    with open(path, 'ab') as f:
        f.write(_DATA.tostring())
    _check(_dl(path))


@responses.activate
def test_download_file(tempdir, mock_urls):
    path = op.join(tempdir, 'test')
    param, url_data, url_checksum = mock_urls
    data_here, data_valid, checksum_here, checksum_valid = param

    assert_succeeds = (data_here and data_valid and
                       ((checksum_here == checksum_valid) or
                        (not(checksum_here) and checksum_valid)))

    download_succeeds = (assert_succeeds or (data_here and
                         (not(data_valid) and not(checksum_here))))

    if download_succeeds:
        data = _dl(path)
    else:
        with raises(Exception):
            data = _dl(path)

    if assert_succeeds:
        _check(data)


@responses.activate
def test_download_sample_data(tempdir):
    name = 'hybrid_10sec.dat'
    url = _BASE_URL['cortexlab'] + name

    _add_mock_response(url, _DATA.tostring())
    _add_mock_response(url + '.md5', _CHECKSUM)

    download_sample_data(name, tempdir)
    with open(op.join(tempdir, name), 'rb') as f:
        data = f.read()
    ae(np.fromstring(data, np.float32), _DATA)

    responses.reset()


@responses.activate
def test_dat_file(tempdir):
    data = np.random.randint(size=(20000, 4),
                             low=-100, high=100).astype(np.int16)
    fn = 'test-4ch-1s.dat'
    _add_mock_response(_BASE_URL['github'] + 'test/' + fn,
                       data.tostring())

    path = download_test_data(fn, tempdir)
    with open(path, 'rb') as f:
        arr = np.fromfile(f, dtype=np.int16).reshape((-1, 4))
    assert arr.shape == (20000, 4)

    responses.reset()



# -*- coding: utf-8 -*-

"""py.test utilities."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
import numpy as np
import os
import os.path as op

from pytest import yield_fixture

from klusta import add_default_handler
from klusta.datasets import download_test_data
from klusta.kwik.mea import load_probe
from klusta.kwik.mock import artificial_traces
from klusta.tempdir import TemporaryDirectory
from klusta.utils import Bunch, _read_python


#------------------------------------------------------------------------------
# Common fixtures
#------------------------------------------------------------------------------

logging.getLogger().setLevel(logging.DEBUG)
add_default_handler('DEBUG')

# Fix the random seed in the tests.
np.random.seed(2016)


@yield_fixture
def tempdir():
    with TemporaryDirectory() as tempdir:
        yield tempdir


@yield_fixture
def chdir_tempdir():
    curdir = os.getcwd()
    with TemporaryDirectory() as tempdir:
        os.chdir(tempdir)
        yield tempdir
    os.chdir(curdir)


@yield_fixture(params=['null', 'artificial', 'real'])
def raw_dataset(request):
    sample_rate = 20000
    curdir = op.realpath(op.dirname(__file__))
    path = op.join(curdir, 'traces/default_settings.py')
    params = _read_python(path)['spikedetekt']
    data_type = request.param
    if data_type == 'real':
        path = download_test_data('test-32ch-10s.dat')
        traces = np.fromfile(path, dtype=np.int16).reshape((200000, 32))
        traces = traces[:45000]
        n_samples, n_channels = traces.shape
        params['use_single_threshold'] = False
        probe = load_probe('1x32_buzsaki')
    else:
        probe = {'channel_groups': {
            0: {'channels': [0, 1, 2],
                'graph': [[0, 1], [0, 2], [1, 2]],
                },
            1: {'channels': [3],
                'graph': [],
                'geometry': {3: [0., 0.]},
                }
        }}
        if data_type == 'null':
            n_samples, n_channels = 25000, 4
            traces = np.zeros((n_samples, n_channels))
        elif data_type == 'artificial':
            n_samples, n_channels = 25000, 4
            traces = artificial_traces(n_samples, n_channels)
            traces[5000:5010, 1] *= 5
            traces[15000:15010, 3] *= 5
    n_samples_w = params['extract_s_before'] + params['extract_s_after']
    yield Bunch(n_channels=n_channels,
                n_samples=n_samples,
                sample_rate=sample_rate,
                n_samples_waveforms=n_samples_w,
                traces=traces,
                params=params,
                probe=probe,
                )



# -*- coding: utf-8 -*-

"""Fixtures."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op

import numpy as np
from pytest import fixture

from klusta.kwik.mea import staggered_positions
from klusta.kwik.mock import (artificial_waveforms,
                              artificial_features,
                              artificial_spike_clusters,
                              artificial_spike_samples,
                              artificial_masks,
                              artificial_traces,
                              )
from klusta.utils import Bunch, _spikes_per_cluster


#------------------------------------------------------------------------------
# Fixtures
#------------------------------------------------------------------------------

@fixture
def model(tempdir):
    model = Bunch()

    n_spikes = 51
    n_samples_w = 31
    n_samples_t = 20000
    n_channels = 11
    n_clusters = 3
    n_features = 4

    model.path = op.join(tempdir, 'test')
    model.n_channels = n_channels
    # TODO: test with permutation and dead channels
    model.channel_order = None
    model.n_spikes = n_spikes
    model.sample_rate = 20000.
    model.duration = n_samples_t / float(model.sample_rate)
    model.spike_times = artificial_spike_samples(n_spikes) * 1.
    model.spike_times /= model.spike_times[-1]
    model.spike_clusters = artificial_spike_clusters(n_spikes, n_clusters)
    model.cluster_ids = np.unique(model.spike_clusters)
    model.channel_positions = staggered_positions(n_channels)
    model.all_waveforms = artificial_waveforms(n_spikes, n_samples_w,
                                               n_channels)
    model.all_masks = artificial_masks(n_spikes, n_channels)
    model.all_traces = artificial_traces(n_samples_t, n_channels)
    model.all_features = artificial_features(n_spikes, n_channels, n_features)

    # features_masks array
    f = model.all_features.reshape((n_spikes, -1))
    m = np.repeat(model.all_masks, n_features, axis=1)
    model.all_features_masks = np.dstack((f, m))

    model.spikes_per_cluster = _spikes_per_cluster(model.spike_clusters)
    model.n_features_per_channel = n_features
    model.n_samples_waveforms = n_samples_w
    model.cluster_groups = {c: None for c in range(n_clusters)}

    return model



# -*- coding: utf-8 -*-

"""Tests of Kwik file opening routines."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op

import numpy as np
from numpy.testing import assert_array_equal as ae
from pytest import raises

from ...utils import captured_logging
from ..creator import create_kwik
from ..mea import MEA, staggered_positions
from ..mock import create_mock_kwik
from ..model import (KwikModel,
                     _list_channel_groups,
                     _list_channels,
                     _list_recordings,
                     _list_clusterings,
                     _concatenate_spikes,
                     )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

_N_CLUSTERS = 10
_N_SPIKES = 100
_N_CHANNELS = 28
_N_FETS = 2
_N_SAMPLES_TRACES = 10000


def test_kwik_utility(tempdir):

    channels = list(range(_N_CHANNELS))

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES)
    model = KwikModel(filename)

    model._kwik.open()
    assert _list_channel_groups(model._kwik.h5py_file) == [1]
    assert _list_recordings(model._kwik.h5py_file) == [0, 1]
    assert _list_clusterings(model._kwik.h5py_file, 1) == ['main',
                                                           'original',
                                                           ]
    assert _list_channels(model._kwik.h5py_file, 1) == channels


def test_concatenate_spikes():
    spikes = [2, 3, 5, 0, 11, 1]
    recs = [0, 0, 0, 1, 1, 2]
    offsets = [0, 7, 100]
    concat = _concatenate_spikes(spikes, recs, offsets)
    ae(concat, [2, 3, 5, 7, 18, 101])


def test_kwik_empty(tempdir):

    channels = [0, 3, 1]
    graph = [[0, 3], [1, 0]]
    probe = {'channel_groups': {
             0: {'channels': channels,
                 'graph': graph,
                 'geometry': {0: (10, 10)},
                 }}}
    sample_rate = 20000

    kwik_path = op.join(tempdir, 'test.kwik')
    create_kwik(kwik_path=kwik_path, probe=probe, sample_rate=sample_rate)

    model = KwikModel(kwik_path)
    ae(model.channels, sorted(channels))
    ae(model.channel_order, channels)

    assert model.sample_rate == sample_rate
    assert model.n_channels == 3
    assert model.spike_samples is None
    assert model.n_spikes == 0
    assert model.n_clusters == 0
    model.describe()


def test_kwik_open_full(tempdir):

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES)

    with raises(ValueError):
        KwikModel()

    # NOTE: n_channels - 2 because we use a special channel order.
    nc = _N_CHANNELS - 2

    # Test implicit open() method.
    kwik = KwikModel(filename)
    kwik.describe()

    kwik.metadata
    ae(kwik.channels, np.arange(_N_CHANNELS))
    assert kwik.n_channels == _N_CHANNELS
    assert kwik.n_spikes == _N_SPIKES
    ae(kwik.channel_order, np.arange(1, _N_CHANNELS - 1)[::-1])

    assert kwik.spike_samples.shape == (_N_SPIKES,)
    assert kwik.spike_samples.dtype == np.int64

    # Make sure the spike samples are increasing, even with multiple
    # recordings.
    # WARNING: need to cast to int64, otherwise negative values will
    # overflow and be positive, making the test pass while the
    # spike samples are *not* increasing!
    assert np.all(np.diff(kwik.spike_samples.astype(np.int64)) >= 0)

    assert kwik.spike_times.shape == (_N_SPIKES,)
    assert kwik.spike_times.dtype == np.float64

    assert kwik.spike_recordings.shape == (_N_SPIKES,)
    assert kwik.spike_recordings.dtype == np.uint16

    assert kwik.spike_clusters.shape == (_N_SPIKES,)
    assert kwik.spike_clusters.min() in (0, 1, 2)
    assert kwik.spike_clusters.max() in(_N_CLUSTERS - 2, _N_CLUSTERS - 1)

    assert kwik.all_features.shape == (_N_SPIKES, nc, _N_FETS)
    kwik.all_features[0, ...]

    assert kwik.all_masks.shape == (_N_SPIKES, nc)

    assert kwik.all_traces.shape == (_N_SAMPLES_TRACES, _N_CHANNELS)

    assert kwik.all_waveforms[0].shape == (1, 40, nc)
    assert kwik.all_waveforms[-1].shape == (1, 40, nc)
    assert kwik.all_waveforms[-10].shape == (1, 40, nc)
    assert kwik.all_waveforms[10].shape == (1, 40, nc)
    assert kwik.all_waveforms[[10, 20]].shape == (2, 40, nc)
    with raises(IndexError):
        kwik.all_waveforms[_N_SPIKES + 10]

    with raises(ValueError):
        kwik.clustering = 'foo'
    with raises(ValueError):
        kwik.channel_group = 42
    assert kwik.n_recordings == 2

    # Test probe.
    assert isinstance(kwik.probe, MEA)
    assert kwik.probe.positions.shape == (nc, 2)
    ae(kwik.probe.positions, staggered_positions(_N_CHANNELS)[1:-1][::-1])

    kwik.close()


def test_kwik_open_no_kwx(tempdir):

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES,
                                with_kwx=False)

    # Test implicit open() method.
    kwik = KwikModel(filename)
    kwik.close()


def test_kwik_open_no_kwd(tempdir):

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES,
                                with_kwd=False)

    # Test implicit open() method.
    kwik = KwikModel(filename)
    with captured_logging() as buf:
        kwik.all_waveforms[:]

    # Enusure that there is no error message.
    assert not buf.getvalue().strip()
    kwik.close()


def test_kwik_save(tempdir):

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES)

    kwik = KwikModel(filename)

    cluster_groups = kwik.cluster_groups
    sc_0 = kwik.spike_clusters.copy()
    sc_1 = sc_0.copy()
    new_cluster = _N_CLUSTERS + 10
    sc_1[_N_SPIKES // 2:] = new_cluster
    ae(kwik.spike_clusters, sc_0)

    kwik.add_cluster_group(4, 'new')
    cluster_groups[new_cluster] = 'new'
    assert kwik.cluster_metadata[new_cluster] == 'unsorted'

    kwik.save(sc_1, cluster_groups, {'test': (1, 2.)})
    ae(kwik.spike_clusters, sc_1)
    assert kwik.cluster_metadata[new_cluster] == 'new'
    kwik.close()

    kwik = KwikModel(filename)
    ae(kwik.spike_clusters, sc_1)
    assert kwik.cluster_metadata[new_cluster] == 'new'
    ae(kwik.clustering_metadata['test'], [1, 2])


def test_kwik_clusterings(tempdir):

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES)

    kwik = KwikModel(filename)
    assert kwik.clusterings == ['main', 'original']

    # The default clustering is 'main'.
    assert kwik.n_spikes == _N_SPIKES
    assert kwik.n_clusters == _N_CLUSTERS
    ae(kwik.cluster_ids, np.arange(_N_CLUSTERS))

    # Change clustering.
    kwik.clustering = 'original'
    n_clu = kwik.n_clusters
    assert kwik.n_spikes == _N_SPIKES
    # Some clusters may be empty with a small number of spikes like here
    assert _N_CLUSTERS * 2 - 4 <= n_clu <= _N_CLUSTERS * 2
    assert len(kwik.cluster_ids) == n_clu


def test_kwik_manage_clusterings(tempdir):

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES)

    kwik = KwikModel(filename)
    spike_clusters = kwik.spike_clusters
    assert kwik.clusterings == ['main', 'original']

    # Test renaming.
    kwik.clustering = 'original'
    with raises(ValueError):
        kwik.rename_clustering('a', 'b')
    with raises(ValueError):
        kwik.rename_clustering('original', 'b')
    with raises(ValueError):
        kwik.rename_clustering('main', 'original')

    kwik.clustering = 'main'
    kwik.rename_clustering('original', 'original_2')
    assert kwik.clusterings == ['main', 'original_2']
    with raises(ValueError):
        kwik.clustering = 'original'
    kwik.clustering = 'original_2'
    n_clu = kwik.n_clusters
    assert len(kwik.cluster_ids) == n_clu

    # Test copy.
    with raises(ValueError):
        kwik.copy_clustering('a', 'b')
    with raises(ValueError):
        kwik.copy_clustering('original', 'b')
    with raises(ValueError):
        kwik.copy_clustering('main', 'original_2')

    # You cannot move the current clustering, but you can copy it.
    with raises(ValueError):
        kwik.rename_clustering('original_2', 'original_2_copy')
    kwik.copy_clustering('original_2', 'original_2_copy')
    kwik.delete_clustering('original_2_copy')

    kwik.clustering = 'main'
    kwik.copy_clustering('original_2', 'original')
    assert kwik.clusterings == ['main', 'original', 'original_2']

    kwik.clustering = 'original'
    ci = kwik.cluster_ids

    kwik.clustering = 'original_2'
    ae(kwik.cluster_ids, ci)

    # Test delete.
    with raises(ValueError):
        kwik.delete_clustering('a')
        kwik.delete_clustering('original')
    kwik.clustering = 'main'
    kwik.delete_clustering('original_2')
    assert kwik.clusterings == ['main', 'original']

    # Test add.
    sc = np.ones(_N_SPIKES, dtype=np.int32)
    sc[1] = sc[-2] = 3
    kwik.add_clustering('new', sc)
    ae(kwik.spike_clusters, spike_clusters)
    kwik.clustering = 'new'
    ae(kwik.spike_clusters, sc)
    assert kwik.n_clusters == 2
    ae(kwik.cluster_ids, [1, 3])


def test_kwik_manage_cluster_groups(tempdir):

    # Create the test HDF5 file in the temporary directory.
    filename = create_mock_kwik(tempdir,
                                n_clusters=_N_CLUSTERS,
                                n_spikes=_N_SPIKES,
                                n_channels=_N_CHANNELS,
                                n_features_per_channel=_N_FETS,
                                n_samples_traces=_N_SAMPLES_TRACES)

    kwik = KwikModel(filename)

    with raises(ValueError):
        kwik.delete_cluster_group(2)
    with raises(ValueError):
        kwik.add_cluster_group(1, 'new')
    with raises(ValueError):
        kwik.rename_cluster_group(1, 'renamed')

    kwik.add_cluster_group(4, 'new')
    kwik.rename_cluster_group(4, 'renamed')

    kwik.delete_cluster_group(4)
    with raises(ValueError):
        kwik.delete_cluster_group(4)




# -*- coding: utf-8 -*-

"""Tests of waveform filtering routines."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae

from ..filter import bandpass_filter, apply_filter, Filter, Whitening


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_apply_filter():
    """Test bandpass filtering on a combination of two sinusoids."""

    rate = 10000.
    low, high = 100., 200.

    # Create a signal with small and high frequencies.
    t = np.linspace(0., 1., rate)
    x = np.sin(2 * np.pi * low / 2 * t) + np.cos(2 * np.pi * high * 2 * t)

    # Filter the signal.
    filter = bandpass_filter(low=low, high=high, order=4, rate=rate)
    ae(apply_filter([], filter=filter), [])

    for x_filtered in (apply_filter(x, filter=filter),
                       Filter(rate=rate, low=low, high=high, order=4)(x),
                       ):

        # Check that the bandpass-filtered signal is weak.
        k = int(2. / low * rate)
        assert np.abs(x[k:-k]).max() >= .9
        assert np.abs(x_filtered[k:-k]).max() <= .1


def test_whitening():
    x = np.random.uniform(size=(100, 10), low=0., high=1.)
    x[:, 1] += .25 * x[:, 0]
    x[:, 5] += .5 * x[:, 0]

    w = Whitening()
    w.fit(x)
    y = w.transform(x)

    assert y.shape == x.shape



# -*- coding: utf-8 -*-

"""Tests of Kwik file creator."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os.path as op

import numpy as np
from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_allclose as ac
from pytest import raises

from ..h5 import open_h5
from ..creator import KwikCreator, _write_by_chunk, create_kwik
from ..mock import (artificial_spike_samples,
                    artificial_features,
                    artificial_masks,
                    )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_write_by_chunk(tempdir):
    n = 5
    arrs = [np.random.rand(i + 1, 3).astype(np.float32) for i in range(n)]
    n_tot = sum(_.shape[0] for _ in arrs)

    path = op.join(tempdir, 'test.h5')
    with open_h5(path, 'w') as f:
        ds = f.write('/test', shape=(n_tot, 3), dtype=np.float32)
        _write_by_chunk(ds, arrs)
    with open_h5(path, 'r') as f:
        ds = f.read('/test')[...]
        offset = 0
        for i, arr in enumerate(arrs):
            size = arr.shape[0]
            assert size == (i + 1)
            ac(ds[offset:offset + size, ...], arr)
            offset += size


def test_creator_simple(tempdir):
    basename = op.join(tempdir, 'my_file')

    creator = KwikCreator(basename)

    # Test create empty files.
    creator.create_empty()
    assert op.exists(basename + '.kwik')
    assert op.exists(basename + '.kwx')

    # Test metadata.
    creator.set_metadata('/application_data/spikedetekt',
                         a=1, b=2., c=[0, 1])

    with open_h5(creator.kwik_path, 'r') as f:
        assert f.read_attr('/application_data/spikedetekt', 'a') == 1
        assert f.read_attr('/application_data/spikedetekt', 'b') == 2.
        ae(f.read_attr('/application_data/spikedetekt', 'c'), [0, 1])

    # Test add spikes in one block.
    n_spikes = 100
    n_channels = 8
    n_features = 3

    spike_samples = artificial_spike_samples(n_spikes)
    features = artificial_features(n_spikes, n_channels, n_features)
    masks = artificial_masks(n_spikes, n_channels)

    creator.add_spikes(group=0,
                       spike_samples=spike_samples,
                       features=features.astype(np.float32),
                       masks=masks.astype(np.float32),
                       n_channels=n_channels,
                       n_features=n_features,
                       )

    # Test the spike samples.
    with open_h5(creator.kwik_path, 'r') as f:
        s = f.read('/channel_groups/0/spikes/time_samples')[...]
        assert s.dtype == np.uint64
        ac(s, spike_samples)

    # Test the features and masks.
    with open_h5(creator.kwx_path, 'r') as f:
        fm = f.read('/channel_groups/0/features_masks')[...]
        assert fm.dtype == np.float32
        ac(fm[:, :, 0], features.reshape((-1, n_channels * n_features)))
        ac(fm[:, ::n_features, 1], masks)

    # Spikes can only been added once.
    with raises(RuntimeError):
        creator.add_spikes(group=0,
                           spike_samples=spike_samples,
                           n_channels=n_channels,
                           n_features=n_features)


def test_creator_chunks(tempdir):
    basename = op.join(tempdir, 'my_file')

    creator = KwikCreator(basename)
    creator.create_empty()

    # Test add spikes in one block.
    n_spikes = 100
    n_channels = 8
    n_features = 3

    spike_samples = artificial_spike_samples(n_spikes)
    features = artificial_features(n_spikes, n_channels,
                                   n_features).astype(np.float32)
    masks = artificial_masks(n_spikes, n_channels).astype(np.float32)

    def _split(arr):
        n = n_spikes // 10
        return [arr[k:k + n, ...] for k in range(0, n_spikes, n)]

    creator.add_spikes(group=0,
                       spike_samples=spike_samples,
                       features=_split(features),
                       masks=_split(masks),
                       n_channels=n_channels,
                       n_features=n_features,
                       )

    # Test the spike samples.
    with open_h5(creator.kwik_path, 'r') as f:
        s = f.read('/channel_groups/0/spikes/time_samples')[...]
        assert s.dtype == np.uint64
        ac(s, spike_samples)

    # Test the features and masks.
    with open_h5(creator.kwx_path, 'r') as f:
        fm = f.read('/channel_groups/0/features_masks')[...]
        assert fm.dtype == np.float32
        ac(fm[:, :, 0], features.reshape((-1, n_channels * n_features)))
        ac(fm[:, ::n_features, 1], masks)


def test_creator_add_no_spikes(tempdir):
    basename = op.join(tempdir, 'my_file')

    creator = KwikCreator(basename)
    creator.create_empty()

    creator.add_spikes(group=0, n_channels=4, n_features=2)


def test_creator_metadata(tempdir):
    basename = op.join(tempdir, 'my_file')

    creator = KwikCreator(basename)

    # Add recording.
    creator.add_recording(0, start_sample=20000, sample_rate=10000)

    with open_h5(creator.kwik_path, 'r') as f:
        assert f.read_attr('/recordings/0', 'start_sample') == 20000
        assert f.read_attr('/recordings/0', 'start_time') == 2.
        assert f.read_attr('/recordings/0', 'sample_rate') == 10000

    # Add probe.
    channels = [0, 3, 1]
    graph = [[0, 3], [1, 0]]
    probe = {'channel_groups': {
             0: {'channels': channels,
                 'graph': graph,
                 'geometry': {0: (10, 10)},
                 }}}
    creator.set_probe(probe)

    with open_h5(creator.kwik_path, 'r') as f:
        ae(f.read_attr('/channel_groups/0', 'channel_order'), channels)
        ae(f.read_attr('/channel_groups/0', 'adjacency_graph'), graph)
        for channel in channels:
            path = '/channel_groups/0/channels/{:d}'.format(channel)
            position = (10, 10) if channel == 0 else (0, channel)
            ae(f.read_attr(path, 'position'), position)

    # Add clustering.
    creator.add_clustering(group=0,
                           name='main',
                           spike_clusters=np.arange(10),
                           cluster_groups={3: 1},
                           )

    with open_h5(creator.kwik_path, 'r') as f:
        sc = f.read('/channel_groups/0/spikes/clusters/main')
        ae(sc, np.arange(10))

        for cluster in range(10):
            path = '/channel_groups/0/clusters/main/{:d}'.format(cluster)
            cg = f.read_attr(path, 'cluster_group')
            assert cg == 3 if cluster != 3 else 1


def test_create_kwik(tempdir):

    channels = [0, 3, 1]
    graph = [[0, 3], [1, 0]]
    probe = {'channel_groups': {
             0: {'channels': channels,
                 'graph': graph,
                 'geometry': {0: (10, 10)},
                 }}}

    kwik_path = op.join(tempdir, 'test.kwik')
    create_kwik(kwik_path=kwik_path,
                probe=probe,
                sample_rate=20000,
                )



# -*- coding: utf-8 -*-
# flake8: noqa

"""Installation script."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os
import os.path as op
import re

from setuptools import setup


#------------------------------------------------------------------------------
# Setup
#------------------------------------------------------------------------------

def _package_tree(pkgroot):
    path = op.dirname(__file__)
    subdirs = [op.relpath(i[0], path).replace(op.sep, '.')
               for i in os.walk(op.join(path, pkgroot))
               if '__init__.py' in i[2]]
    return subdirs


curdir = op.dirname(op.realpath(__file__))
readme = open(op.join(curdir, 'README.md')).read()


# Find version number from `__init__.py` without executing it.
filename = op.join(curdir, 'klusta/__init__.py')
with open(filename, 'r') as f:
    version = re.search(r"__version__ = '([^']+)'", f.read()).group(1)


setup(
    name='klusta',
    version=version,
    license="BSD",
    description='Spike detection and automatic clustering for spike sorting',
    long_description=readme,
    author='Kwik Team',
    author_email='cyrille.rossant at gmail.com',
    url='https://klusta.cortexlab.net',
    packages=_package_tree('klusta'),
    package_dir={'klusta': 'klusta'},
    package_data={
        'klusta': ['*.txt', '*.prb'],
    },
    entry_points={
        'console_scripts': [
            'klusta = klusta.launch:main'
        ],
    },
    include_package_data=True,
    keywords='klusta,neuroscience,spike sorting,klustakwik',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



# -*- coding: utf-8 -*-

"""SpikeDetekt algorithm."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from collections import defaultdict
import logging

import numpy as np
from tqdm import tqdm

from ..utils import (Bunch,
                     get_excerpts,
                     chunk_bounds,
                     data_chunk,
                     _concatenate,
                     )
from ..kwik.mea import (_channels_per_group,
                        _probe_adjacency_list,
                        )
from .detect import Thresholder, compute_threshold, FloodFillDetector
from .filter import Filter
from .pca import PCA
from .store import SpikeDetektStore
from .waveform import WaveformExtractor

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Utils
#------------------------------------------------------------------------------

def _keep_spikes(samples, bounds):
    """Only keep spikes within the bounds `bounds=(start, end)`."""
    start, end = bounds
    return (start <= samples) & (samples <= end)


def _split_spikes(groups, idx=None, **arrs):
    """Split spike data according to the channel group."""
    # split: {group: {'spike_samples': ..., 'waveforms':, 'masks':}}
    dtypes = {'spike_samples': np.float64,
              'waveforms': np.float32,
              'masks': np.float32,
              }
    groups = np.asarray(groups)
    if idx is not None:
        assert idx.dtype == np.bool
        n_spikes_chunk = np.sum(idx)
        # First, remove the overlapping bands.
        groups = groups[idx]
        arrs_bis = arrs.copy()
        for key, arr in arrs.items():
            arrs_bis[key] = arr[idx]
            assert len(arrs_bis[key]) == n_spikes_chunk
    # Then, split along the group.
    groups_u = np.unique(groups)
    out = {}
    for group in groups_u:
        i = (groups == group)
        out[group] = {}
        for key, arr in arrs_bis.items():
            out[group][key] = _concat(arr[i], dtypes.get(key, None))
    return out


def _array_list(arrs):
    out = np.empty((len(arrs),), dtype=np.object)
    out[:] = arrs
    return out


def _concat(arr, dtype=None):
    out = np.array([_[...] for _ in arr], dtype=dtype)
    return out


def _cut_traces(traces, interval_samples):
    n_samples, n_channels = traces.shape
    # Take a subset if necessary.
    if interval_samples is not None:
        start, end = interval_samples
    else:
        return traces, 0
    assert 0 <= start < end
    # WARNING: this loads all traces into memory! To fix this properly,
    # we'll have to implement lazy chunking in ConcatenatedTraces.
    size = (end - start) * traces.shape[1] * traces.dtype.itemsize / 1024. ** 3
    if size > 1:
        logger.warn("Loading all traces in memory: this will require %.3f GB "
                    "of RAM! ", size)
        logger.warn("To avoid this, do not specify `--interval`. "
                    "This bug will be fixed later.")
    traces = traces[start:end, ...]
    if start > 0:
        # TODO: add offset to the spike samples...
        raise NotImplementedError("Need to add `start` to the "
                                  "spike samples")
    return traces, start


def _subtract_offsets(samples, offsets):
    """Subtract the recording offsets from spike samples.

    Return the subtracted spike samples and the spike_recordings array.

    """
    if samples is None:
        return None, None
    assert isinstance(samples, (list, np.ndarray))
    samples = np.asarray(samples).copy()
    assert len(offsets) >= 2
    assert offsets[0] == 0
    assert offsets[-1] >= samples[-1]
    n = len(offsets) - 1

    # Find where to insert the offsets in the spike samples.
    ind = np.searchsorted(samples, offsets)
    assert ind[0] == 0
    assert len(ind) == n + 1

    spike_recordings = np.zeros(len(samples), dtype=np.int32)

    # Loop over all recordings.
    for rec in range(n):
        # Start of the current recording.
        start_rec = offsets[rec]
        # Spike indices of the first spikes in every recording.
        i, j = ind[rec], ind[rec + 1]
        # Ensure that the selected spikes belong to the current recording.
        if i < len(samples) - 1:
            assert start_rec <= samples[i]
        # Subtract the current recording offset to the selected spikes.
        samples[i:j] -= start_rec
        # Create the spike_recordings array.
        spike_recordings[i:j] = rec
        assert np.all(samples[i:j] >= 0)
        assert np.all(samples[i:j] <= (offsets[rec + 1] - start_rec))

    return samples, spike_recordings


def _relative_channels(channels, adjacency):
    """Renumber channels from absolute indices to relative indices,
    to match arrays used in the detection code.

    Parameters
    ----------

    channels : dict
        A dict {group: list_of_channels}
    adjacency : dict
        A dict {group: set_of_neighbors}

    """
    ch_out = {}
    adj_out = {}
    mapping = {}
    offset = 0
    for group in channels:
        ch = channels[group]
        n = len(ch)
        ch_out[group] = [i + offset for i in range(n)]
        # From absolute to relative indices.
        mapping.update({c: (i + offset) for i, c in enumerate(ch)})
        offset += n
    # Recreate the adjacency dict by applying the mapping to
    # both the keys and values.
    for c, i in mapping.items():
        adj_out[i] = set(mapping[_] for _ in adjacency.get(c, set())
                         if _ in mapping)
    return ch_out, adj_out


#------------------------------------------------------------------------------
# Spike detection class
#------------------------------------------------------------------------------

class SpikeDetekt(object):
    """Spike detection class.

    Parameters
    ----------

    tempdir : str
        Path to the temporary directory used by the algorithm. It should be
        on a SSD for best performance.
    probe : dict
        The probe dictionary.
    **kwargs : dict
        Spike detection parameters.

    """
    def __init__(self, tempdir=None, probe=None, **kwargs):
        super(SpikeDetekt, self).__init__()
        self._tempdir = tempdir
        # Load a probe.
        if probe is not None:
            kwargs['probe_channels'] = _channels_per_group(probe)
            kwargs['probe_adjacency_list'] = _probe_adjacency_list(probe)
        self._kwargs = kwargs
        self._n_channels_per_group = {
            group: len(channels)
            for group, channels in self._kwargs['probe_channels'].items()
        }
        for group in self._n_channels_per_group:
            logger.info("Found %d live channels in group %d.",
                        self._n_channels_per_group[group], group)
        self._groups = sorted(self._n_channels_per_group)
        self._n_features = self._kwargs['n_features_per_channel']
        before = self._kwargs['extract_s_before']
        after = self._kwargs['extract_s_after']
        self._n_samples_waveforms = before + after

    # Processing objects creation
    # -------------------------------------------------------------------------

    def _create_filter(self):
        rate = self._kwargs['sample_rate']
        low = self._kwargs['filter_low']
        high = self._kwargs['filter_high_factor'] * rate
        order = self._kwargs['filter_butter_order']
        return Filter(rate=rate,
                      low=low,
                      high=high,
                      order=order,
                      )

    def _create_thresholder(self, thresholds=None):
        mode = self._kwargs['detect_spikes']
        return Thresholder(mode=mode, thresholds=thresholds)

    def _create_detector(self):
        graph = self._kwargs['probe_adjacency_list']
        probe_channels = self._kwargs['probe_channels']
        join_size = self._kwargs['connected_component_join_size']
        return FloodFillDetector(probe_adjacency_list=graph,
                                 join_size=join_size,
                                 channels_per_group=probe_channels,
                                 )

    def _create_extractor(self, thresholds):
        before = self._kwargs['extract_s_before']
        after = self._kwargs['extract_s_after']
        weight_power = self._kwargs['weight_power']
        probe_channels = self._kwargs['probe_channels']
        return WaveformExtractor(extract_before=before,
                                 extract_after=after,
                                 weight_power=weight_power,
                                 channels_per_group=probe_channels,
                                 thresholds=thresholds,
                                 )

    def _create_pca(self):
        n_pcs = self._kwargs['n_features_per_channel']
        return PCA(n_pcs=n_pcs)

    # Misc functions
    # -------------------------------------------------------------------------

    def update_params(self, **kwargs):
        self._kwargs.update(kwargs)

    # Processing functions
    # -------------------------------------------------------------------------

    def apply_filter(self, data):
        """Filter the traces."""
        filter = self._create_filter()
        return filter(data).astype(np.float32)

    def find_thresholds(self, traces):
        """Find weak and strong thresholds in filtered traces."""
        rate = self._kwargs['sample_rate']
        n_excerpts = self._kwargs['n_excerpts']
        excerpt_size = int(self._kwargs['excerpt_size_seconds'] * rate)
        single = bool(self._kwargs['use_single_threshold'])
        strong_f = self._kwargs['threshold_strong_std_factor']
        weak_f = self._kwargs['threshold_weak_std_factor']

        logger.info("Finding the thresholds...")
        excerpt = get_excerpts(traces,
                               n_excerpts=n_excerpts,
                               excerpt_size=excerpt_size)
        excerpt_f = self.apply_filter(excerpt)
        thresholds = compute_threshold(excerpt_f,
                                       single_threshold=single,
                                       std_factor=(weak_f, strong_f))
        logger.debug("Thresholds: {}.".format(thresholds))
        return {'weak': thresholds[0],
                'strong': thresholds[1]}

    def detect(self, traces_f, thresholds=None):
        """Detect connected waveform components in filtered traces.

        Parameters
        ----------

        traces_f : array
            An `(n_samples, n_channels)` array with the filtered data.
        thresholds : dict
            The weak and strong thresholds.

        Returns
        -------

        components : list
            A list of `(n, 2)` arrays with `sample, channel` pairs.

        """
        # Threshold the data following the weak and strong thresholds.
        thresholder = self._create_thresholder(thresholds)
        # Transform the filtered data according to the detection mode.
        traces_t = thresholder.transform(traces_f)
        # Compute the threshold crossings.
        weak = thresholder.detect(traces_t, 'weak')
        strong = thresholder.detect(traces_t, 'strong')

        # Find dead channels.
        cpg = self._kwargs['probe_channels']
        live_channels = sorted([item for sublist in cpg.values()
                                for item in sublist])
        n_channels = traces_f.shape[1]
        dead_channels = np.setdiff1d(np.arange(n_channels), live_channels)
        logger.debug("Dead channels: %s.", ', '.join(map(str, dead_channels)))

        # Kill threshold crossings on dead channels.
        weak[:, dead_channels] = 0
        strong[:, dead_channels] = 0

        # Run the detection.
        detector = self._create_detector()
        return detector(weak_crossings=weak,
                        strong_crossings=strong)

    def extract_spikes(self, components, traces_f,
                       thresholds=None, keep_bounds=None, s_start=None):
        """Extract spikes from connected components.

        Returns a split object.

        Parameters
        ----------
        components : list
            List of connected components.
        traces_f : array
            Filtered data.
        thresholds : dict
            The weak and strong thresholds.
        keep_bounds : tuple
            (keep_start, keep_end).
        s_start : 0
            Start of the chunk.

        """
        n_spikes = len(components)
        if n_spikes == 0:
            return {}

        # Transform the filtered data according to the detection mode.
        thresholder = self._create_thresholder()
        traces_t = thresholder.transform(traces_f)
        # Extract all waveforms.
        extractor = self._create_extractor(thresholds)
        groups, samples, waveforms, masks = zip(*[extractor(component,
                                                            data=traces_f,
                                                            data_t=traces_t,
                                                            )
                                                  for component in components])

        # Create the return arrays.
        groups = np.array(groups, dtype=np.int32)
        assert groups.shape == (n_spikes,)
        assert groups.dtype == np.int32

        samples = np.array(samples, dtype=np.float64)
        assert samples.shape == (n_spikes,)
        assert samples.dtype == np.float64

        # These are lists of arrays of various shapes (because of various
        # groups).
        waveforms = _array_list(waveforms)
        assert waveforms.shape == (n_spikes,)
        assert waveforms.dtype == np.object

        masks = _array_list(masks)
        assert masks.dtype == np.object
        assert masks.shape == (n_spikes,)

        # Reorder the spikes.
        idx = np.argsort(samples)
        groups = groups[idx]
        samples = samples[idx]
        waveforms = waveforms[idx]
        masks = masks[idx]

        # Remove spikes in the overlapping bands.
        # WARNING: add s_start to spike_samples, because spike_samples
        # is relative to the start of the chunk.
        # It is important to add s_start and not keep_start, because of
        # edge effects between overlapping chunks.
        s_start = s_start or 0
        (keep_start, keep_end) = keep_bounds
        idx = _keep_spikes(samples + s_start, (keep_start, keep_end))

        # Split the data according to the channel groups.
        split = _split_spikes(groups, idx=idx, spike_samples=samples,
                              waveforms=waveforms, masks=masks)
        # split: {group: {'spike_samples': ..., 'waveforms':, 'masks':}}

        # Assert that spike samples are increasing.
        for group in split:
            samples = split[group]['spike_samples']
            if samples is not None:
                assert np.all(np.diff(samples) >= 0)

        return split

    def waveform_pcs(self, waveforms, masks):
        """Compute waveform principal components.

        Returns
        -------

        pcs : array
            An `(n_features, n_samples, n_channels)` array.

        """
        pca = self._create_pca()
        if waveforms is None or not len(waveforms):
            return
        assert (waveforms.shape[0], waveforms.shape[2]) == masks.shape
        return pca.fit(waveforms, masks)

    def features(self, waveforms, pcs):
        """Extract features from waveforms.

        Returns
        -------

        features : array
            An `(n_spikes, n_channels, n_features)` array.

        """
        pca = self._create_pca()
        out = pca.transform(waveforms, pcs=pcs)
        assert out.dtype == np.float32
        return out

    # Chunking
    # -------------------------------------------------------------------------

    def iter_chunks(self, n_samples):
        """Iterate over chunks."""
        rate = self._kwargs['sample_rate']
        chunk_size = int(self._kwargs['chunk_size_seconds'] * rate)
        overlap = int(self._kwargs['chunk_overlap_seconds'] * rate)
        for chunk_idx, bounds in enumerate(chunk_bounds(n_samples, chunk_size,
                                                        overlap=overlap)):
            yield Bunch(bounds=bounds,
                        s_start=bounds[0],
                        s_end=bounds[1],
                        keep_start=bounds[2],
                        keep_end=bounds[3],
                        keep_bounds=(bounds[2:4]),
                        key=bounds[2],
                        chunk_idx=chunk_idx,
                        )

    def n_chunks(self, n_samples):
        """Number of chunks."""
        return len(list(self.iter_chunks(n_samples)))

    def chunk_keys(self, n_samples):
        return [chunk.key for chunk in self.iter_chunks(n_samples)]

    # Output data
    # -------------------------------------------------------------------------

    def output_data(self):
        """Bunch of values to be returned by the algorithm."""
        sc = self._store.spike_counts
        chunk_keys = self._store.chunk_keys

        # NOTE: deal with multiple recordings.
        samples = self._store.spike_samples()

        s = {}
        r = {}
        for group in self._groups:
            spikes = _concatenate(samples[group])
            # Check increasing spikes.
            if spikes is not None:
                assert np.all(np.diff(spikes) >= 0)
            s[group], r[group] = _subtract_offsets(spikes,
                                                   self.recording_offsets)

        output = Bunch(groups=self._groups,
                       n_chunks=len(chunk_keys),
                       chunk_keys=chunk_keys,
                       spike_samples=s,  # dict
                       recordings=r,  # dict
                       masks=self._store.masks(),
                       features=self._store.features(),
                       n_channels_per_group=self._n_channels_per_group,
                       n_features_per_channel=self._n_features,
                       spike_counts=sc,
                       n_spikes_total=sc(),
                       n_spikes_per_group={group: sc(group=group)
                                           for group in self._groups},
                       n_spikes_per_chunk={chunk_key: sc(chunk_key=chunk_key)
                                           for chunk_key in chunk_keys},
                       )
        return output

    # Main loop
    # -------------------------------------------------------------------------

    def _iter_spikes(self, n_samples, step_spikes=1, thresholds=None):
        """Iterate over extracted spikes (possibly subset).

        Yield a split dictionary `{group: {'waveforms': ..., ...}}`.

        """
        for chunk in self.iter_chunks(n_samples):

            # Extract a few components.
            components = self._store.load(name='components',
                                          chunk_key=chunk.key)
            if components is None or not len(components):
                yield chunk, {}
                continue

            k = np.clip(step_spikes, 1, len(components))
            components = components[::k]

            # Get the filtered chunk.
            chunk_f = self._store.load(name='filtered',
                                       chunk_key=chunk.key)

            # Extract the spikes from the chunk.
            split = self.extract_spikes(components, chunk_f,
                                        keep_bounds=chunk.keep_bounds,
                                        s_start=chunk.s_start,
                                        thresholds=thresholds,
                                        )

            yield chunk, split

    def step_detect(self, traces=None, thresholds=None):
        n_samples, n_channels = traces.shape
        n_chunks = self.n_chunks(n_samples)

        # Use the recording offsets when dealing with multiple recordings.
        self.recording_offsets = getattr(traces, 'offsets', [0, len(traces)])

        # Pass 1: find the connected components and count the spikes.
        # self._pr.start_step('detect', n_chunks)

        # Dictionary {chunk_key: components}.
        # Every chunk has a unique key: the `keep_start` integer.
        n_spikes_total = 0
        for chunk in tqdm(self.iter_chunks(n_samples),
                          desc='Detecting spikes'.ljust(24),
                          total=n_chunks, leave=True):
            chunk_data = data_chunk(traces, chunk.bounds, with_overlap=True)

            # Apply the filter.
            data_f = self.apply_filter(chunk_data)
            assert data_f.dtype == np.float32
            assert data_f.shape == chunk_data.shape

            # Save the filtered chunk.
            self._store.store(name='filtered', chunk_key=chunk.key,
                              data=data_f)

            # Detect spikes in the filtered chunk.
            components = self.detect(data_f, thresholds=thresholds)
            self._store.store(name='components', chunk_key=chunk.key,
                              data=components)

            # Report progress.
            n_spikes_chunk = len(components)
            n_spikes_total += n_spikes_chunk
            logger.debug("Found %d spikes in chunk %d.", n_spikes_chunk,
                         chunk.key)
            # self._pr.increment(n_spikes=n_spikes_chunk,
            #                    n_spikes_total=n_spikes_total)

        return n_spikes_total

    def step_excerpt(self, n_samples=None,
                     n_spikes_total=None, thresholds=None):
        # self._pr.start_step('excerpt', self.n_chunks(n_samples))
        n_chunks = self.n_chunks(n_samples)
        k = int(n_spikes_total / float(self._kwargs['pca_n_waveforms_max']))
        w_subset = defaultdict(list)
        m_subset = defaultdict(list)
        n_spikes_total = 0
        for chunk, split in tqdm(self._iter_spikes(n_samples, step_spikes=k,
                                                   thresholds=thresholds),
                                 desc='Extracting waveforms'.ljust(24),
                                 total=n_chunks,
                                 leave=True,
                                 ):
            n_spikes_chunk = 0
            for group, out in split.items():
                w_subset[group].append(out['waveforms'])
                m_subset[group].append(out['masks'])
                assert len(out['masks']) == len(out['waveforms'])
                n_spikes_chunk += len(out['masks'])

            n_spikes_total += n_spikes_chunk
            # self._pr.increment(n_spikes=n_spikes_chunk,
            #                    n_spikes_total=n_spikes_total)
        for group in self._groups:
            w_subset[group] = _concatenate(w_subset[group])
            m_subset[group] = _concatenate(m_subset[group])

        return w_subset, m_subset

    def step_pcs(self, w_subset=None, m_subset=None):
        # self._pr.start_step('pca', len(self._groups))
        pcs = {}
        for group in tqdm(self._groups,
                          desc='Performing PCA'.ljust(24),
                          total=len(self._groups),
                          leave=True,
                          ):
            # Perform PCA and return the components.
            pcs[group] = self.waveform_pcs(w_subset[group],
                                           m_subset[group])
            # self._pr.increment()
        return pcs

    def step_extract(self, n_samples=None,
                     pcs=None, thresholds=None):
        # self._pr.start_step('extract', self.n_chunks(n_samples))
        # chunk_counts = defaultdict(dict)  # {group: {key: n_spikes}}.
        # n_spikes_total = 0
        for chunk, split in tqdm(self._iter_spikes(n_samples,
                                                   thresholds=thresholds),
                                 desc='Computing features'.ljust(24),
                                 total=self.n_chunks(n_samples),
                                 leave=True,
                                 ):
            # Delete filtered and components cache files.
            # self._store.delete(name='filtered', chunk_key=chunk.key)
            # self._store.delete(name='components', chunk_key=chunk.key)
            # split: {group: {'spike_samples': ..., 'waveforms':, 'masks':}}
            for group, out in split.items():
                out['features'] = self.features(out['waveforms'], pcs[group])
                # Checking that spikes are increasing.
                spikes = out['spike_samples']
                self._store.append(group=group,
                                   chunk_key=chunk.key,
                                   spike_samples=spikes,
                                   features=out['features'],
                                   masks=out['masks'],
                                   spike_offset=chunk.s_start,
                                   )
            # n_spikes_total = self._store.spike_counts()
            # n_spikes_chunk = self._store.spike_counts(chunk_key=chunk.key)
            # self._pr.increment(n_spikes_total=n_spikes_total,
            #                    n_spikes=n_spikes_chunk)

    def run_serial(self, traces, interval_samples=None):
        """Run SpikeDetekt using one CPU."""
        traces, offset = _cut_traces(traces, interval_samples)
        n_samples, n_channels = traces.shape

        # Initialize the main loop.
        chunk_keys = self.chunk_keys(n_samples)
        # n_chunks = len(chunk_keys)

        # TODO: progress reporter
        class PR(object):
            def start_step(self, *args, **kwargs):
                pass

            def increment(self, *args, **kwargs):
                pass

        # self._pr = PR(n_chunks=n_chunks)
        self._store = SpikeDetektStore(self._tempdir,
                                       groups=self._groups,
                                       chunk_keys=chunk_keys)

        # Find the weak and strong thresholds.
        thresholds = self.find_thresholds(traces)

        # Spike detection.
        n_spikes_total = self.step_detect(traces=traces,
                                          thresholds=thresholds)

        # Excerpt waveforms.
        w_subset, m_subset = self.step_excerpt(n_samples=n_samples,
                                               n_spikes_total=n_spikes_total,
                                               thresholds=thresholds)

        # Compute the PCs.
        pcs = self.step_pcs(w_subset=w_subset, m_subset=m_subset)

        # Compute all features.
        self.step_extract(n_samples=n_samples, pcs=pcs, thresholds=thresholds)

        return self.output_data()



# -*- coding: utf-8 -*-

"""Test MEA."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

from pytest import raises
import numpy as np
from numpy.testing import assert_array_equal as ae

from ..mea import (_probe_channels, _probe_positions, _probe_adjacency_list,
                   MEA, linear_positions, staggered_positions,
                   load_probe, list_probes
                   )


#------------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------------

def test_probe():
    probe = {'channel_groups': {
             0: {'channels': [0, 3, 1],
                 'graph': [[0, 3], [1, 0]],
                 'geometry': {0: (10, 10), 1: (10, 20), 3: (20, 30)},
                 },
             1: {'channels': [7],
                 'graph': [],
                 },
             }}
    adjacency = {0: set([1, 3]),
                 1: set([0]),
                 3: set([0]),
                 }
    assert _probe_channels(probe, 0) == [0, 3, 1]
    ae(_probe_positions(probe, 0), [(10, 10), (20, 30), (10, 20)])
    assert _probe_adjacency_list(probe) == adjacency

    mea = MEA(probe=probe)
    assert mea.positions is None
    assert mea.channels is None
    assert mea.n_channels == 0
    assert mea.adjacency == adjacency
    assert mea.channels_per_group == {0: [0, 3, 1], 1: [7]}

    mea.change_channel_group(0)
    ae(mea.positions, [(10, 10), (20, 30), (10, 20)])
    assert mea.channels == [0, 3, 1]
    assert mea.n_channels == 3


def test_mea():

    n_channels = 10
    channels = np.arange(n_channels)
    positions = np.random.randn(n_channels, 2)

    mea = MEA(channels)
    mea.positions = positions
    ae(mea.positions, positions)
    assert mea.adjacency is None

    mea = MEA(channels, positions=positions)
    assert mea.n_channels == n_channels

    mea = MEA(channels, positions=positions)
    assert mea.n_channels == n_channels

    with raises(AssertionError):
        MEA(channels=np.arange(n_channels + 1), positions=positions)

    with raises(AssertionError):
        MEA(channels=channels, positions=positions[:-1, :])

    mea = MEA(channels=channels)
    assert mea.n_channels == n_channels
    mea.positions = positions
    with raises(ValueError):
        mea.positions = positions[:-1, :]


def test_positions():
    probe = staggered_positions(31)
    assert probe.shape == (31, 2)
    ae(probe[-1], (0, 0))

    probe = linear_positions(29)
    assert probe.shape == (29, 2)


def test_library():
    probe = load_probe('1x32_buzsaki')
    assert probe
    assert probe['channel_groups'][0]['channels'] == list(range(32))

    assert '1x32_buzsaki' in list_probes()



# -*- coding: utf-8 -*-
# flake8: noqa

"""The KlustaSuite: automatic spike sorting up to 64 channels."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
import os
import os.path as op
from pkg_resources import get_distribution, DistributionNotFound
import subprocess


#------------------------------------------------------------------------------
# Global variables and functions
#------------------------------------------------------------------------------

def _git_version():
    curdir = os.getcwd()
    filedir, _ = op.split(__file__)
    os.chdir(filedir)
    try:
        fnull = open(os.devnull, 'w')
        version = ('-git-' + subprocess.check_output(
                   ['git', 'describe', '--abbrev=8', '--dirty',
                    '--always', '--tags'],
                   stderr=fnull).strip().decode('ascii'))
        return version
    except (OSError, subprocess.CalledProcessError):
        return ""
    finally:
        os.chdir(curdir)


__author__ = 'KwikTeam'
__email__ = 'cyrille.rossant at gmail.com'
__version__ = '3.0.16'
__version_git__ = __version__ + _git_version()


# Set a null handler on the root logger
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.NullHandler())


_logger_fmt = '%(asctime)s [%(levelname)s] %(caller)s %(message)s'
_logger_date_fmt = '%H:%M:%S'


class _Formatter(logging.Formatter):
    def format(self, record):
        # Only keep the first character in the level name.
        record.levelname = record.levelname[0]
        filename = op.splitext(op.basename(record.pathname))[0]
        record.caller = '{:s}:{:d}'.format(filename, record.lineno).ljust(20)
        return super(_Formatter, self).format(record)


def add_default_handler(level='INFO'):
    handler = logging.StreamHandler()
    handler.setLevel(level)

    formatter = _Formatter(fmt=_logger_fmt,
                           datefmt=_logger_date_fmt)
    handler.setFormatter(formatter)

    logger.addHandler(handler)



# -*- coding: utf-8 -*-

"""Tests of waveform loader."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_allclose as ac
import numpy.random as npr
from pytest import raises

from ...kwik.mock import artificial_traces
from ..waveform import _slice, WaveformLoader, WaveformExtractor
from ..filter import bandpass_filter, apply_filter


#------------------------------------------------------------------------------
# Tests extractor
#------------------------------------------------------------------------------

def test_extract_simple():
    weak = 1.
    strong = 2.
    nc = 4
    ns = 20
    channels = list(range(nc))
    cpg = {0: channels}
    # graph = {0: [1, 2], 1: [0, 2], 2: [0, 1], 3: []}

    data = np.random.uniform(size=(ns, nc), low=0., high=1.)

    data[10, 0] = 0.5
    data[11, 0] = 1.5
    data[12, 0] = 1.0

    data[10, 1] = 1.5
    data[11, 1] = 2.5
    data[12, 1] = 2.0

    component = np.array([[10, 0],
                          [10, 1],
                          [11, 0],
                          [11, 1],
                          [12, 0],
                          [12, 1],
                          ])

    we = WaveformExtractor(extract_before=3,
                           extract_after=5,
                           thresholds={'weak': weak,
                                       'strong': strong},
                           channels_per_group=cpg,
                           )

    # _component()
    comp = we._component(component, n_samples=ns)
    ae(comp.comp_s, [10, 10, 11, 11, 12, 12])
    ae(comp.comp_ch, [0, 1, 0, 1, 0, 1])
    assert (comp.s_min, comp.s_max) == (10 - 3, 12 + 4)
    ae(comp.channels, range(nc))

    # _normalize()
    assert we._normalize(weak) == 0
    assert we._normalize(strong) == 1
    ae(we._normalize([(weak + strong) / 2.]), [.5])

    # _comp_wave()
    wave = we._comp_wave(data, comp)
    assert wave.shape == (3 + 5 + 1, nc)
    ae(wave[3:6, :], [[0.5, 1.5, 0., 0.],
                      [1.5, 2.5, 0., 0.],
                      [1.0, 2.0, 0., 0.]])

    # masks()
    masks = we.masks(data, wave, comp)
    ae(masks, [.5, 1., 0, 0])

    # spike_sample_aligned()
    s = we.spike_sample_aligned(wave, comp)
    assert 11 <= s < 12

    # extract()
    wave_e = we.extract(data, s, channels=channels)
    assert wave_e.shape[1] == wave.shape[1]
    ae(wave[3:6, :2], wave_e[3:6, :2])

    # align()
    wave_a = we.align(wave_e, s)
    assert wave_a.shape == (3 + 5, nc)

    # Test final call.
    groups, s_f, wave_f, masks_f = we(component, data=data, data_t=data)
    assert s_f == s
    assert np.all(groups == 0)
    ae(masks_f, masks)
    ac(wave_f, wave_a)

    # Tests with a different order.
    we = WaveformExtractor(extract_before=3,
                           extract_after=5,
                           thresholds={'weak': weak,
                                       'strong': strong},
                           channels_per_group={0: [1, 0, 3]},
                           )
    groups, s_f_o, wave_f_o, masks_f_o = we(component, data=data, data_t=data)
    assert np.all(groups == 0)
    assert s_f == s_f_o
    assert np.allclose(wave_f[:, [1, 0, 3]], wave_f_o)
    ac(masks_f_o, [1., 0.5, 0.])


#------------------------------------------------------------------------------
# Tests loader
#------------------------------------------------------------------------------

def test_slice():
    assert _slice(0, (20, 20)) == slice(0, 20, None)


def test_loader():
    n_samples_trace, n_channels = 10000, 100
    n_samples = 40
    n_spikes = n_samples_trace // (2 * n_samples)

    traces = artificial_traces(n_samples_trace, n_channels)
    spike_samples = np.cumsum(npr.randint(low=0, high=2 * n_samples,
                                          size=n_spikes))

    with raises(ValueError):
        WaveformLoader(traces)

    # Create a loader.
    loader = WaveformLoader(traces, n_samples=n_samples)
    assert id(loader.traces) == id(traces)
    loader.traces = traces

    # Extract a waveform.
    t = spike_samples[10]
    waveform = loader._load_at(t)
    assert waveform.shape == (n_samples, n_channels)
    ae(waveform, traces[t - 20:t + 20, :])

    waveforms = loader[spike_samples[10:20]]
    assert waveforms.shape == (10, n_samples, n_channels)
    t = spike_samples[15]
    w1 = waveforms[5, ...]
    w2 = traces[t - 20:t + 20, :]
    assert np.allclose(w1, w2)


def test_edges():
    n_samples_trace, n_channels = 1000, 10
    n_samples = 40

    traces = artificial_traces(n_samples_trace, n_channels)

    # Filter.
    b_filter = bandpass_filter(rate=1000,
                               low=50,
                               high=200,
                               order=3)
    filter_margin = 10

    # Create a loader.
    loader = WaveformLoader(traces,
                            n_samples=n_samples,
                            filter=lambda x: apply_filter(x, b_filter),
                            filter_margin=filter_margin)

    # Invalid time.
    with raises(ValueError):
        loader._load_at(200000)

    assert loader._load_at(0).shape == (n_samples, n_channels)
    assert loader._load_at(5).shape == (n_samples, n_channels)
    assert loader._load_at(n_samples_trace - 5).shape == (n_samples,
                                                          n_channels)
    assert loader._load_at(n_samples_trace - 1).shape == (n_samples,
                                                          n_channels)


def test_loader_channels():
    n_samples_trace, n_channels = 1000, 50
    n_samples = 40

    traces = artificial_traces(n_samples_trace, n_channels)

    # Create a loader.
    loader = WaveformLoader(traces, n_samples=n_samples)
    loader.traces = traces
    channels = [10, 20, 30]
    loader.channels = channels
    assert loader.channels == channels
    assert loader[500].shape == (1, n_samples, 3)
    assert loader[[500, 501, 600, 300]].shape == (4, n_samples, 3)

    # Test edge effects.
    assert loader[3].shape == (1, n_samples, 3)
    assert loader[995].shape == (1, n_samples, 3)

    with raises(NotImplementedError):
        loader[500:510]


def test_loader_filter():
    n_samples_trace, n_channels = 1000, 100
    n_samples = 40
    n_spikes = n_samples_trace // (2 * n_samples)

    traces = artificial_traces(n_samples_trace, n_channels)
    spike_samples = np.cumsum(npr.randint(low=0, high=2 * n_samples,
                                          size=n_spikes))

    # With filter.
    def my_filter(x):
        return x * x

    loader = WaveformLoader(traces,
                            n_samples=(n_samples // 2, n_samples // 2),
                            filter=my_filter,
                            filter_margin=5)

    t = spike_samples[5]
    waveform_filtered = loader._load_at(t)
    traces_filtered = my_filter(traces)
    traces_filtered[t - 20:t + 20, :]
    assert np.allclose(waveform_filtered, traces_filtered[t - 20:t + 20, :])



# -*- coding: utf-8 -*-

"""Tests of spike detection routines."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import numpy as np
from numpy.testing import assert_array_equal as ae
from numpy.testing import assert_allclose as ac

from ..detect import (compute_threshold,
                      Thresholder,
                      connected_components,
                      FloodFillDetector,
                      )
from ...kwik.mock import artificial_traces


#------------------------------------------------------------------------------
# Test thresholder
#------------------------------------------------------------------------------

def test_compute_threshold():
    n_samples, n_channels = 100, 10
    data = artificial_traces(n_samples, n_channels)

    # Single threshold.
    threshold = compute_threshold(data, std_factor=1.)
    assert threshold.shape == (2,)
    assert threshold[0] > 0
    assert threshold[0] == threshold[1]

    threshold = compute_threshold(data, std_factor=[1., 2.])
    assert threshold.shape == (2,)
    assert threshold[1] == 2 * threshold[0]

    # Multiple threshold.
    threshold = compute_threshold(data, single_threshold=False, std_factor=2.)
    assert threshold.shape == (2, n_channels)

    threshold = compute_threshold(data,
                                  single_threshold=False,
                                  std_factor=(1., 2.))
    assert threshold.shape == (2, n_channels)
    ac(threshold[1], 2 * threshold[0])


def test_thresholder():
    n_samples, n_channels = 100, 12
    strong, weak = .1, .2

    data = artificial_traces(n_samples, n_channels)

    # Positive and strong.
    thresholder = Thresholder(mode='positive',
                              thresholds=strong)
    ae(thresholder(data), data > strong)

    # Negative and weak.
    thresholder = Thresholder(mode='negative',
                              thresholds={'weak': weak})
    ae(thresholder(data), data < -weak)

    # Both and strong+weak.
    thresholder = Thresholder(mode='both',
                              thresholds={'weak': weak,
                                          'strong': strong,
                                          })
    ae(thresholder(data, 'weak'), np.abs(data) > weak)
    ae(thresholder(data, threshold='strong'), np.abs(data) > strong)

    # Multiple thresholds.
    t = thresholder(data, ('weak', 'strong'))
    ae(t['weak'], np.abs(data) > weak)
    ae(t['strong'], np.abs(data) > strong)

    # Array threshold.
    thre = np.linspace(weak - .05, strong + .05, n_channels)
    thresholder = Thresholder(mode='positive', thresholds=thre)
    t = thresholder(data)
    assert t.shape == data.shape
    ae(t, data > thre)


#------------------------------------------------------------------------------
# Test connected components
#------------------------------------------------------------------------------

def _as_set(c):
    if isinstance(c, np.ndarray):
        c = c.tolist()
    c = [tuple(_) for _ in c]
    return set(c)


def _assert_components_equal(cc1, cc2):
    assert len(cc1) == len(cc2)
    for c1, c2 in zip(cc1, cc2):
        assert _as_set(c1) == _as_set(c2)


def _test_components(chunk=None, components=None, **kwargs):

    def _clip(x, m, M):
        return [_ for _ in x if m <= _ < M]

    n = 5
    probe_adjacency_list = {i: set(_clip([i - 1, i + 1], 0, n))
                            for i in range(n)}

    if chunk is None:
        chunk = [[0, 0, 0, 0, 0],
                 [0, 0, 1, 0, 0],
                 [1, 0, 1, 1, 0],
                 [1, 0, 0, 1, 0],
                 [0, 1, 0, 1, 1],
                 ]

    if components is None:
        components = []

    if not isinstance(chunk, np.ndarray):
        chunk = np.array(chunk)
    strong_crossings = kwargs.pop('strong_crossings', None)
    if (strong_crossings is not None and
            not isinstance(strong_crossings, np.ndarray)):
        strong_crossings = np.array(strong_crossings)

    comp = connected_components(chunk,
                                probe_adjacency_list=probe_adjacency_list,
                                strong_crossings=strong_crossings,
                                **kwargs)
    _assert_components_equal(comp, components)


def test_components():
    # 1 time step, 1 element
    _test_components([[0, 0, 0, 0, 0]], [])

    _test_components([[1, 0, 0, 0, 0]], [[(0, 0)]])

    _test_components([[0, 1, 0, 0, 0]], [[(0, 1)]])

    _test_components([[0, 0, 0, 1, 0]], [[(0, 3)]])

    _test_components([[0, 0, 0, 0, 1]], [[(0, 4)]])

    # 1 time step, 2 elements
    _test_components([[1, 1, 0, 0, 0]], [[(0, 0), (0, 1)]])

    _test_components([[1, 0, 1, 0, 0]], [[(0, 0)], [(0, 2)]])

    _test_components([[1, 0, 0, 0, 1]], [[(0, 0)], [(0, 4)]])

    _test_components([[0, 1, 0, 1, 0]], [[(0, 1)], [(0, 3)]])

    # 1 time step, 3 elements
    _test_components([[1, 1, 1, 0, 0]], [[(0, 0), (0, 1), (0, 2)]])

    _test_components([[1, 1, 0, 1, 0]], [[(0, 0), (0, 1)], [(0, 3)]])

    _test_components([[1, 0, 1, 1, 0]], [[(0, 0)], [(0, 2), (0, 3)]])

    _test_components([[0, 1, 1, 1, 0]], [[(0, 1), (0, 2), (0, 3)]])

    _test_components([[0, 1, 1, 0, 1]], [[(0, 1), (0, 2)], [(0, 4)]])

    # 5 time steps, varying join_size
    _test_components([
        [0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0],
        [1, 0, 1, 1, 0],
        [1, 0, 0, 1, 0],
        [0, 1, 0, 1, 1],
    ], [[(1, 2)],
        [(2, 0)],
        [(2, 2), (2, 3)],
        [(3, 0)],
        [(3, 3)],
        [(4, 1)],
        [(4, 3), (4, 4)],
        ])

    _test_components([
        [0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0],
        [1, 0, 1, 1, 0],
        [1, 0, 0, 1, 0],
        [0, 1, 0, 1, 1],
    ], [[(1, 2), (2, 2), (2, 3), (3, 3), (4, 3), (4, 4)],
        [(2, 0), (3, 0), (4, 1)]], join_size=1)

    _test_components(
        components=[[(1, 2), (2, 2), (2, 3), (3, 3), (4, 3), (4, 4),
                     (2, 0), (3, 0), (4, 1)]], join_size=2)

    # 5 time steps, strong != weak
    _test_components(join_size=0,
                     strong_crossings=[
                         [0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0]])

    _test_components(components=[[(1, 2)]],
                     join_size=0,
                     strong_crossings=[
                         [0, 0, 0, 0, 0],
                         [0, 0, 1, 0, 0],
                         [0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0]])

    _test_components(
        components=[[(1, 2), (2, 2), (2, 3), (3, 3), (4, 3), (4, 4)]],
        join_size=1,
        strong_crossings=[
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1]])

    _test_components(
        components=[[(1, 2), (2, 2), (2, 3), (3, 3), (4, 3), (4, 4),
                     (2, 0), (3, 0), (4, 1)]],
        join_size=2,
        strong_crossings=[
            [0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0]])

    _test_components(
        components=[[(1, 2), (2, 2), (2, 3), (3, 3), (4, 3), (4, 4),
                     (2, 0), (3, 0), (4, 1)]],
        join_size=2,
        strong_crossings=[
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0]])


def test_flood_fill():

    graph = {0: [1, 2], 1: [0, 2], 2: [0, 1], 3: []}

    channels = {0: [1, 2, 3]}

    ff = FloodFillDetector(probe_adjacency_list=graph,
                           join_size=1,
                           channels_per_group=channels
                           )

    weak = [[0, 0, 0, 0],
            [0, 1, 1, 0],
            [0, 0, 0, 0],
            [0, 0, 1, 0],
            [0, 0, 1, 1],
            [0, 0, 0, 0],
            [0, 0, 0, 1],
            [0, 0, 0, 0],
            ]

    # Weak - weak
    comps = [[(1, 1), (1, 2)],
             [(3, 2), (4, 2)],
             [(4, 3)],
             [(6, 3)],
             ]
    cc = ff(weak, weak)
    _assert_components_equal(cc, comps)

    # Weak and strong
    strong = [[0, 0, 0, 0],
              [0, 0, 0, 0],
              [0, 0, 0, 0],
              [0, 0, 1, 0],
              [0, 0, 0, 1],
              [0, 0, 0, 0],
              [0, 0, 0, 1],
              [0, 0, 0, 0],
              ]

    comps = [[(3, 2), (4, 2)],
             [(4, 3)],
             [(6, 3)],
             ]
    cc = ff(weak, strong)
    _assert_components_equal(cc, comps)

    channels = {0: [1, 2, 3, 4]}

    ff = FloodFillDetector(probe_adjacency_list=graph,
                           join_size=2,
                           channels_per_group=channels
                           )

    weak = [[0, 0, 0, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0],
            ]

    comps = [[(1, 1)],
             [(1, 3)],
             [(4, 4), (6, 4)],
             ]

    cc = ff(weak, weak)
    _assert_components_equal(cc, comps)



# -*- coding: utf-8 -*-

"""Launch routines."""

#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import logging
from operator import itemgetter
import os.path as op
import shutil
import sys

import click
import numpy as np

from .traces import SpikeDetekt
from .kwik.creator import create_kwik, KwikCreator
from .kwik.model import KwikModel
from .klustakwik import klustakwik
from .utils import _ensure_dir_exists, _read_python
from .__init__ import __version_git__, add_default_handler

logger = logging.getLogger(__name__)


#------------------------------------------------------------------------------
# Launch
#------------------------------------------------------------------------------

def detect(model, interval=None, **kwargs):
    traces = model.all_traces

    # Setup the temporary directory.
    expdir = op.dirname(model.kwik_path)
    sd_dir = op.join(expdir, '.spikedetekt')
    _ensure_dir_exists(sd_dir)

    # Default interval.
    if interval is not None:
        (start_sec, end_sec) = interval
        sr = model.sample_rate
        interval_samples = (int(start_sec * sr),
                            int(end_sec * sr))
    else:
        interval_samples = None

    # Take the parameters in the Kwik file, coming from the PRM file.
    params = model.metadata
    params.update(kwargs)
    # Pretty print params.
    logger.info("SpikeDetekt parameters:")
    for key, value in params.items():
        logger.info("%s = %s", key, value)

    # Probe parameters required by SpikeDetekt.
    params['probe_channels'] = model.probe.channels_per_group
    params['probe_adjacency_list'] = model.probe.adjacency
    assert 'probe' not in params

    # Start the spike detection.
    logger.debug("Running SpikeDetekt...")
    sd = SpikeDetekt(tempdir=sd_dir, **params)
    out = sd.run_serial(traces, interval_samples=interval_samples)
    return out


def cluster(model, spike_ids=None, tempdir=None, **kwargs):
    """Return the spike_clusters and metadata.

    Doesn't make any change to the model. The caller must add the clustering.

    """
    # Skip if no spikes.
    if not model.n_spikes:
        return np.array([], dtype=np.int32), {}

    # Setup the temporary directory.
    expdir = op.dirname(model.kwik_path)
    kk_dir = tempdir or op.join(expdir, '.klustakwik2')
    _ensure_dir_exists(kk_dir)

    params = model.kk2_metadata
    params.update(kwargs)

    # Original spike_clusters array.
    if model.spike_clusters is None:
        n_spikes = (len(spike_ids) if spike_ids is not None
                    else model.n_spikes)
        spike_clusters_orig = np.zeros(n_spikes, dtype=np.int32)
    else:
        spike_clusters_orig = model.spike_clusters.copy()

    # Save clustering checkpoint to a text file.
    path = op.join(kk_dir, 'spike_clusters.%d.txt' % model.channel_group)

    def on_iter(sc):
        # Update the original spike clusters.
        spike_clusters = spike_clusters_orig.copy()
        spike_clusters[spike_ids] = sc
        # Backup.
        if op.exists(path):
            shutil.copy(path, path + '~')
        np.savetxt(path, spike_clusters, fmt='%d')

    # Load initial clusters.
    spike_clusters = None
    if op.exists(path):
        spike_clusters = np.loadtxt(path).astype(np.int64)
        if spike_clusters.shape != (model.n_spikes,):
            logger.warn("Unable to resume KK from `%s`, there are %d values "
                        "instead of %d", path, len(spike_clusters), n_spikes)
            spike_clusters = None
        else:
            logger.info("Resuming KK from `%s` (%d clusters)...",
                        path,
                        len(np.unique(spike_clusters)))
            logger.info("If you want to restart the clustering, delete `%s`.",
                        path)

    if spike_clusters is None:
        logger.info("Starting KK...")

    # Run KK.
    sc, params = klustakwik(model=model,
                            spike_ids=spike_ids,
                            spike_clusters=spike_clusters,
                            iter_callback=on_iter,
                            **params)
    logger.info("The automatic clustering process has finished.")

    # Save the results in the Kwik file.
    spike_clusters = spike_clusters_orig.copy()
    spike_clusters[spike_ids] = sc

    # Set the new clustering metadata.
    metadata = {'klustakwik2_{}'.format(name): value
                for name, value in params.items()}

    return sc, metadata


def save_clu(spike_clusters, clu_path):
    arr = np.concatenate(([len(np.unique(spike_clusters))], spike_clusters))
    np.savetxt(clu_path, arr, fmt='%d', newline='\n')


def save_res(spike_samples, res_path):
    np.savetxt(res_path, spike_samples, fmt='%d', newline='\n')


def klusta(prm_file,
           output_dir=None,
           interval=None,
           channel_group=None,
           detect_only=False,
           cluster_only=False,
           overwrite=False,
           legacy_output=None,
           ):

    if prm_file.endswith('.kwik'):
        kwik_path = prm_file
        logger.info("Call `klusta` on a PRM file to spikesort your data.")
        if op.exists(kwik_path):
            logger.info("Here is a description of `%s`:", kwik_path)
            model = KwikModel(kwik_path)
            model.describe()
        return

    # Detection and/or clustering.
    do_detect = not cluster_only
    do_cluster = not detect_only
    assert do_detect or do_cluster

    # Parse the interval (pair of floats).
    # if isinstance(interval, six.string_types):
    if interval and interval[0] is not None and interval[1] is not None:
        assert 0 <= interval[0] < interval[1]
    else:
        interval = None

    # Delete the SpikeDetekt cache if overwriting the files.
    if overwrite:
        for temp_dir in ('.spikedetekt', '.klustakwik2', '.phy'):
            sd_dir = op.join(output_dir or '', temp_dir)
            if op.exists(sd_dir):
                logger.info("Deleting `%s`.", sd_dir)
                shutil.rmtree(sd_dir)

    # Find the path to the kwik file.
    prm = _read_python(prm_file) if prm_file else {}
    output_dir = output_dir or ''
    kwik_path = op.join(output_dir, prm['experiment_name'] + '.kwik')

    # Detection.
    if do_detect:

        # Ensure the kwik file exists. Doesn't overwrite it if it exists.
        create_kwik(prm_file=prm_file,
                    output_dir=output_dir,
                    overwrite=overwrite,
                    )

        logger.info("Starting spike detection.")
        # NOTE: always detect on all shanks.
        model = KwikModel(kwik_path)
        out = detect(model, interval=interval)
        model.close()

        # Add the spikes to the kwik file.
        creator = KwikCreator(kwik_path)
        creator.add_spikes_after_detection(out)
        ns = ', '.join('%d spikes found in shank %d' % (n, i)
                       for (i, n) in sorted(out.n_spikes_per_group.items(),
                                            key=itemgetter(0)))
        logger.info("Spike detection done! " + ns)

        # Save res files.
        if legacy_output:
            for group in out.groups:
                res_path = '%s.%d.res' % (prm['experiment_name'], group)
                res_path = op.join(output_dir, res_path)
                logger.info("Save %d spikes to `%s`.",
                            len(out.spike_samples[group]), res_path)
                save_res(out.spike_samples[group], res_path)

    # List of channel groups.
    if channel_group is None:
        # If no channel group specified, find the list of all channel groups.
        model = KwikModel(kwik_path)
        channel_groups = model.channel_groups
        model.close()
    else:
        channel_groups = [channel_group]

    # Clustering.
    if do_cluster:

        # Update KlustaKwik2 parameters.
        creator = KwikCreator(kwik_path)
        creator.set_metadata('/application_data/klustakwik2',
                             **prm.get('klustakwik2', {}))

        # Cluster every channel group.
        for channel_group in channel_groups:
            logger.info("Starting clustering on shank %d/%d.",
                        channel_group, len(channel_groups))
            model = KwikModel(kwik_path, channel_group=channel_group)
            logger.info("Clustering group %d (%d spikes).",
                        channel_group, model.n_spikes)
            # Skip clustering if there are no spikes.
            if not model.n_spikes:
                continue
            spike_clusters, metadata = cluster(model)
            model.close()

            # Save clu files.
            if legacy_output:
                clu_path = '%s.%d.clu' % (prm['experiment_name'],
                                          channel_group)
                clu_path = op.join(output_dir, clu_path)
                logger.info("Save %d spikes to `%s`.",
                            len(spike_clusters), clu_path)
                save_clu(spike_clusters, clu_path)

            logger.info("Save the clustering to `%s`.", kwik_path)
            # Or add the results to the kwik file.
            model = KwikModel(kwik_path, channel_group=channel_group)

            # Backup and remove the main clustering.
            if 'main' in model.clusterings:
                i = len([_ for _ in model.clusterings
                         if _.startswith('backup')])
                name = 'backup-%d' % i
                model.copy_clustering('main', name)
                model.clustering = name
                model.delete_clustering('main')
                model.add_clustering('main', spike_clusters)
            else:
                model.add_clustering('main', spike_clusters)
                model.copy_clustering('main', 'original')

            # Switch to main.
            model.clustering = 'main'
            model.clustering_metadata.update(metadata)
            model.save(clustering_metadata=model.clustering_metadata)
            model.close()
        logger.info("Clustering done!")

    return kwik_path


@click.command()
@click.argument('prm_file',
                type=click.Path(exists=True, file_okay=True, dir_okay=False),
                )
@click.option('--output-dir',
              type=click.Path(file_okay=False, dir_okay=True),
              help='Output directory.',
              )
@click.option('--interval',
              type=click.Tuple([float, float]),
              help='Interval in seconds, e.g. `--interval 0 2`.',
              default=(None, None),
              )
@click.option('--channel-group',
              type=click.INT,
              help='Channel group to cluster (all by default).',
              )
@click.option('--detect-only',
              help='Only do spike detection.',
              default=False,
              is_flag=True,
              )
@click.option('--cluster-only',
              help='Only do automatic clustering.',
              default=False,
              is_flag=True,
              )
@click.option('--legacy-output',
              help=('Save `.res` and `.clu` files in addition to `.kwik`.'),
              default=False,
              is_flag=True,
              )
@click.option('--overwrite',
              help='Overwrite the Kwik file.',
              default=False,
              is_flag=True,
              )
@click.option('--debug',
              help='Use the DEBUG logging level.',
              default=False,
              is_flag=True,
              )
@click.version_option(version=__version_git__)
@click.help_option()
def main(*args, **kwargs):
    """Spikesort a dataset.

    By default, perform spike detection (with SpikeDetekt) and automatic
    clustering (with KlustaKwik2). You can also choose to run only one step.

    You need to specify three pieces of information to spikesort your data:

    * The raw data file: typically a `.dat` file.

    * The PRM file: a Python file with the `.prm` extension, containing the parameters for your sorting session.

    * The PRB file: a Python file with the `.prb` extension, containing the layout of your probe.

    """  # noqa

    debug = kwargs.pop('debug', None)

    # Hide the traceback unless DEBUG mode.
    def exception_handler(exception_type, exception, traceback):
        print("{}: {}".format(exception_type.__name__, exception))
    if not debug:
        sys.excepthook = exception_handler

    add_default_handler('DEBUG' if debug else 'INFO')
    return klusta(*args, **kwargs)
